import sys
from pyneval.model.swc_node import SwcNode, SwcTree
from pyneval.io import swc_writer

SWC_PATH = "../../data/optimation/gold.swc"
OUTPUT_PATH = "../../data/optimation/test4_gold.swc"


def cut_swc_rectangle(swc_tree, LFD_pos, RBT_pos):
    '''
    swc_tree: input the swc tree you want to cut
    LFD_pos: 1*3 tuple, refer to Left Front Down position of the cut rectangle
    RBT_pos: 1*3 tuple, refer to Right Behind Top position of the cut rectangle
    require:
    LFD_pos[0] < RBT_pos[0]
    LFD_pos[1] < RBT_pos[1]
    LFD_pos[2] < RBT_pos[2]
    the output region will be saved in (pyneal)/output/swc_cut/
    '''
    if LFD_pos[0] > RBT_pos[0] or LFD_pos[1] > RBT_pos[1] or LFD_pos[2] > RBT_pos[2]:
        raise Exception("[Error: ]LFD pos should be smaller than RBT pos")
    for node in swc_tree.get_node_list():
        if node.is_virtual():
            continue
        if not (LFD_pos[0] <= node.get_x() <= RBT_pos[0] and \
                LFD_pos[1] <= node.get_y() <= RBT_pos[1] and \
                LFD_pos[2] <= node.get_z() <= RBT_pos[2]):
            swc_tree.remove_node(node)
                            

if __name__ == "__main__":
    sys.setrecursionlimit(100000)
    swc_tree = SwcTree()
    # load origin swc file
    swc_tree.load(SWC_PATH)

    cut_swc_rectangle(swc_tree, (260, 30, 125), (324, 94, 189))
    swc_tree.get_node_list(update=True)
    for node in swc_tree.get_node_list():
        node.set_x(node.get_x()-260)
        node.set_y(node.get_y()-30)
        node.set_z(node.get_z()-125)
    # the result will be saved in:
    swc_writer.swc_save(swc_tree=swc_tree, out_path=OUTPUT_PATH)



import re
import seaborn as sns
import matplotlib.pyplot as plt
import matplotlib.cm as cm
from matplotlib.pyplot import MultipleLocator, LinearLocator
from matplotlib import rcParams
import shutil
import os
import csv
import math
sns.set_style('whitegrid')


def read_data(file_path):
    xs, ys = [], []
    fg = 0
    with open(file_path, 'r') as f:
        lines = f.readlines()
        for line in lines:
            if line[0] == '#':
                continue
            print(line.split())
            v = list(map(float, line.split()))
            x_y = list()
            for item in v:
                x_y.append(item)
            if fg == 0:
                xs.append(x_y)
            else:
                ys.append(x_y)
            fg = 1 - fg
    return xs, ys


def linechar_visual(avgs, stds, output_dir, lables=None, file_name=None):
    it = ["{}%".format(i*10) for i in range(len(avgs[0]))]

    colors = ['#FF3030', '#FF8C00', '#4876FF', '#458B00']
    plt.figure(figsize=(3, 3))
    ax = plt.subplot(335)
    ax.grid(False)
    mmax = -0x3fffffff
    mmin = 0x3fffffff
    for i in range(len(avgs)):
        ax.plot(it, avgs[i], color=colors[i], label=lables[i])
        r1 = list(map(lambda x: x[0]-x[1]*2, zip(avgs[i], stds[i])))
        r2 = list(map(lambda x: x[0]+x[1]*2, zip(avgs[i], stds[i])))
        r3 = list(map(lambda x: x[0]-x[1], zip(avgs[i], stds[i])))
        r4 = list(map(lambda x: x[0]+x[1], zip(avgs[i], stds[i])))
        ax.fill_between(it, r1, r2, color=colors[i], alpha=0.1)
        ax.fill_between(it, r3, r4, color=colors[i], alpha=0.3)

        mmax = max(mmax, max(avgs[i]))
        mmin = min(mmin, min(avgs[i]))
        # ax.legend()

        # ax.set_xlabel('moved node number', fontsize=30)
        # ax.set_ylabel('score', fontsize=30)
    plt.xlim(0, 10)
    mmin = round(mmin, 1)
    mmax = round(mmax, 1)
    plt.ylim(float('%.2g' % mmin), float('%.2g' % mmax))

    x_major_locator = MultipleLocator(10)
    y_major_locator = LinearLocator(3)

    ax.spines["top"].set_color("none")
    ax.spines["right"].set_color("none")
    ax.spines["left"].set_color("black")
    ax.spines["bottom"].set_color("black")
    ax.spines["left"].set_linewidth(1.5)
    ax.spines["bottom"].set_linewidth(1.5)

    ax.xaxis.set_major_locator(x_major_locator)
    ax.yaxis.set_major_locator(y_major_locator)
    plt.xticks(fontsize=10, weight="bold")
    plt.yticks(fontsize=10, weight="bold")
    # plt.show()
    plt.savefig(os.path.join(output_dir, file_name), dpi=100)


# def plot_one_metric_all_data():
#     plt.figure(figsize=(6, 1), dpi=135)
#     # plt.figure(1)
#     ax1 = plt.subplot(161)
#     plt.plot([1, 2, 3, 4], [4, 5, 7, 8], color="r", linestyle="--")
#
#     ax2 = plt.subplot(162)
#     plt.plot([1, 2, 3, 4], [4, 5, 7, 8], color="r", linestyle="--")
#
#     ax3 = plt.subplot(163)
#     plt.plot([1, 2, 3, 4], [4, 5, 7, 8], color="r", linestyle="--")
#
#     ax4 = plt.subplot(164)
#     plt.plot([1, 2, 3, 4], [4, 5, 7, 8], color="r", linestyle="--")
#
#     ax5 = plt.subplot(165)
#     plt.plot([1, 2, 3, 4], [4, 5, 7, 8], color="r", linestyle="--")
#
#     ax6 = plt.subplot(166)
#     plt.plot([1, 2, 3, 4], [4, 5, 7, 8], color="r", linestyle="--")
#
#     ax2.set_yticklabels([])
#     ax3.set_yticklabels([])
#     ax4.set_yticklabels([])
#     ax5.set_yticklabels([])
#     ax6.set_yticklabels([])
#
#     ax1.grid(False)
#     ax2.grid(False)
#     ax3.grid(False)
#     ax4.grid(False)
#     ax5.grid(False)
#     ax6.grid(False)
#
#     ax1.set_xlim(0, )
#
#     plt.show()


if __name__ == "__main__":
    # # avgs, stds = read_data(".\input_data\lm_delete_recall.txt")
    # # lables = ['diadem_example', '30_18_10_gold', '34_23_10']
    # # output_dir = "../../../output/csv_dir"
    # # fig_name = 'lm_delete_precition.png'
    # # # linechar_visual(avgs, stds, lables=['2_18','30_18_10','34_23_10'], file_name='lm_delete_recall.png')
    # # linechar_visual(avgs, stds, lables=lables, file_name=fig_name, )
    #
    # # root csv fold for input files
    # data_dir = "../../../data/csv_data"
    # # fold for output png
    # output_dir = "../../../output/csv_dir"
    #
    # # select result on different swc files
    # for file_name in os.listdir(data_dir):
    #     file_dir = os.path.join(data_dir, file_name)
    #     # select result on different metric method
    #     for method in os.listdir(file_dir):
    #         method_dir = os.path.join(file_dir, method)
    #         # there are two ways to generate random data: delete and move_5
    #         delete_dir = os.path.join(method_dir, "delete")
    #         move_dir = os.path.join(method_dir, "fake_data.txt")
    #         stds = []
    #         avgs = []
    #         with open(move_dir) as f:
    #             reader = csv.reader(f)
    #             fg = 0
    #             for row in reader:
    #                 row = list(map(float, row))
    #                 if fg == 0:
    #                     avgs.append(row)
    #                 else:
    #                     stds.append(row)
    #                 fg ^= 1
    #
    #         linechar_visual(avgs=avgs, stds=stds, output_dir=output_dir, file_name=file_name+"move_5")
    pass


"""Converts images to TFRecords file format with Example protos."""

import os
import sys
import glob
import json
import tensorflow as tf
from PIL import Image
import tifffile
import numpy as np
import Utils
from Label import class2label

def _int64_feature(value):
    return tf.train.Feature(int64_list=tf.train.Int64List(value=[value]))


def _bytes_feature(value):
    return tf.train.Feature(bytes_list=tf.train.BytesList(value=[value]))


def _readFolders(
    input_path,
    class_label_file_path,
    organize_type,
    resize,
    new_size,
	is_3d,
    number):

    class_2_label = Utils.json2dict(class_label_file_path)
    im_shape = []
    if organize_type==3:
        file_paths = glob.glob(os.path.join(input_path, '*'))
    else:
        file_paths = glob.glob(os.path.join(input_path, '*/*'))

    np.random.shuffle(file_paths)
    try:
        for file_name in file_paths:
            if is_3d:
                im = tifffile.imread(file_name)
            else:
                im = Image.open(file_name)
            if resize:
                im = im.resize(new_size)

            im = np.asarray(im, np.uint8)
            if len(im.shape)==2:
                im = im.reshape(im.shape[0], im.shape[1], 1)
            elif len(im.shape)==3:
                im = im.reshape(im.shape[0], im.shape[1], im.shape[2], 1)
            else:
                continue
            # check shape
            if len(im_shape)>0 and im_shape[0]!=im.shape:
                sys.exit('input image size not same')
            else:
                if len(im_shape)==0:
                    im_shape.append(im.shape)
                else:
                    im_shape[0] = im.shape

            if organize_type==1:
                class_name = os.path.basename(os.path.dirname(file_name))
            elif organize_type==2:
                class_name = os.path.basename(file_name).split('-')[0]
            elif organize_type==3:
                # don't care label, for gan train
                class_name = 'branch'
            else:
                sys.exit('organize type error')

            global result
            if class_name in class_2_label:
            	label = class_2_label[class_name]
            else:
                continue

            if class_name not in result:
                result[class_name] = 1
            else:
                result[class_name] += 1

            yield im, label

    except IOError as e:
        sys.exit('image read error')

    print('summary: input image shape is: ', im_shape[0])


def convertTo(
    input_path_list,
    output_path,
    class_label_file_path,
    organize_type,
	is_3d=False,
	resize=False,
    new_size=None,
    number=None):
    ''' create tfrecords file
        input_path_list (string)
        output_path (string) : tfrecords output file path
        organize_type (int) : 1->organize by folder, 2->organize in one folder, 3->don't care label for gan
        is_3d (bool) : is input image 3d
    '''
    global result
    result = dict()
    writer = tf.python_io.TFRecordWriter(output_path)
    for input_path in input_path_list:
        for image, label in _readFolders(
            input_path,
            class_label_file_path,
            organize_type,
			resize,
			new_size,
            is_3d,
            number):
            image_raw = image.tostring()
            image_shape = image.shape
            if is_3d:
                example = tf.train.Example(
                    features=tf.train.Features(
                        feature={
                            'depth': _int64_feature(int(image_shape[0])),
                            'width': _int64_feature(int(image_shape[1])),
                            'height': _int64_feature(int(image_shape[2])),
                            'channel': _int64_feature(int(image_shape[3])),
                            'label': _int64_feature(int(label)),
                            'image_raw': _bytes_feature(image_raw)
                            }
                        )
                    )
            else:
                example = tf.train.Example(
                    features=tf.train.Features(
                        feature={
                            'height': _int64_feature(image_shape[0]),
                            'width': _int64_feature(image_shape[1]),
                            'depth': _int64_feature(image_shape[2]),
                            'label': _int64_feature(label),
                            'image_raw': _bytes_feature(image_raw)
                            }
                        )
                    )

            writer.write(example.SerializeToString())

    return result

if __name__ == '__main__':
    tf.app.run()



# bennieHan 2019-11-12 16:01
# all right reserved

import math
import os
import queue

import numpy as np
from anytree import NodeMixin, RenderTree, iterators
from pyneval.model.euclidean_point import EuclideanPoint

_3D = "3d"
_2D = "2d"


def Make_Virtual():
    return SwcNode(nid=-1, center=EuclideanPoint(center=[0, 0, 0]))


def get_nearby_swc_node_list(gold_node, test_kdtree, test_pos_node_dict, threshold):
    """
    find all nodes in "test_swc_list" which are close enough to "gold_node"
    sort them by distance

    :param gold_node: swc_node
    :param test_swc_list:
    :param threshold:
    :return:
    """
    if gold_node.is_virtual():
        return
    tmp_list = []
    # find the closest pos for gold node
    target_pos_list = test_kdtree.search_knn(list(gold_node.get_center_as_tuple()), k=5)
    for pos in target_pos_list:
        target_node = test_pos_node_dict[tuple(pos[0].data)]
        if target_node.is_virtual() or gold_node.is_virtual():
            continue
        # only if gold and test nodes are very close(dis < 0.03), they can be considered as the same pos
        if gold_node.distance(target_node) < threshold:
            tmp_list.append(tuple([target_node, target_node.distance(gold_node)]))

    tmp_list.sort(key=lambda x: x[1])
    res_list = []
    for tu in tmp_list:
        res_list.append(tu[0])
    return res_list


# not used old code
def compute_platform_area(r1, r2, h):
    return (r1 + r2) * h * math.pi


# not used old code
def compute_two_node_area(tn1, tn2, remain_dist):
    """Returns the surface area formed by two nodes
    """
    r1 = tn1.radius()
    r2 = tn2.radius()
    d = tn1.distance(tn2)
    print (remain_dist)

    if remain_dist >= d:
        h = d
    else:
        h = remain_dist
        a = remain_dist / d
        r2 = r1 * (1 - a) + r2 * a

    area = compute_platform_area(r1, r2, h)
    return area


# not used old code
def compute_surface_area(tn, range_radius):
    area = 0

    # backtrace
    currentDist = 0
    parent = tn.parent
    while parent and currentDist < range_radius:
        remainDist = range_radius - currentDist
        area += compute_two_node_area(tn, parent, remainDist)
        currentDist += tn.distance(parent)
        tn = parent
        parent = tn.parent

    # forwardtrace
    currentDist = 0
    childList = tn.children
    while len(childList) == 1 and currentDist < range_radius:
        child = childList[0]
        remainDist = range_radius - currentDist
        area += compute_two_node_area(tn, child, remainDist)
        currentDist += tn.distance(child)
        tn = child
        childList = tn.children

    return area


def get_lca(u, v):
    tmp_set = set()
    tmp_u = u
    tmp_v = v
    while u.get_id() != -1:
        tmp_set.add(u.get_id())
        u = u.parent

    while v.get_id() != -1:
        if v.get_id() in tmp_set:
            return v.get_id()
        v = v.parent
    return None


class SwcNode(NodeMixin):
    """
        this is a class that temporarily store SWC file
        Attributes:
        id: id of the node,
        type: leaf = 1,continuation = 2, bifurcation = 3,
        parent: pa node,
        son=[]: son list,
        center: Euclidean point describe node center
        radius: radius of the node
        surface_area: surface area of the cylinder, radius is current radius, length is the distance to its parent
        volume: volume of the cylinder. radious is the same as above

        parent_trajectory: distance to root
        left_trajectory: distance to the farthest son of left_son
        right_trajectory: distance to the farthest son of right_son

        path_length: distance to parent
        xy_path_length: distance to parent regardless z coordinate
        z_path_lenth: distance to parent
    """

    def __init__(
        self,
        nid=-1,
        ntype=0,
        radius=1.0,
        center=None,
        parent=None,
        depth=0,
        surface_area=0.0,
        volume=0.0,
        parent_trajectory=None,
        left_trajectory=None,
        right_trajectory=None,
        route_length=0.0,
        path_length=0.0,
        xy_path_length=0.0,
        z_path_lenth=0.0,
    ):
        self._id = nid
        self._type = ntype
        self.parent = parent
        self._pos = center if center is not None else EuclideanPoint(center=[0, 0, 0])
        self._radius = radius
        self.surface_area = surface_area
        self.volume = volume
        self._depth = depth

        self.parent_trajectory = parent_trajectory
        self.left_trajectory = left_trajectory
        self.right_trajectory = right_trajectory

        self.root_length = route_length
        self.path_length = path_length
        self.xy_path_length = xy_path_length
        self.z_path_length = z_path_lenth

    def set_id(self, id):
        self._id = id

    def get_id(self):
        return self._id

    def get_x(self):
        return self._pos.get_x()

    def get_y(self):
        return self._pos.get_y()

    def get_z(self):
        return self._pos.get_z()

    def set_x(self, x):
        self._pos.set_x(x)

    def set_y(self, y):
        self._pos.set_y(y)

    def set_z(self, z):
        self._pos.set_z(z)

    def set_r(self, r):
        self._radius = r

    def get_center(self):
        return self._pos

    def get_center_as_tuple(self):
        return tuple([round(self.get_x(), 2), round(self.get_y(), 2), round(self.get_z(), 2)])

    def set_center(self, center):
        if not isinstance(center, EuclideanPoint):
            raise Exception("[Error: ]not EuclideanPoint")
        del self._pos
        self._pos = center

    def depth(self):
        return self._depth

    def radius(self):
        return self._radius

    def get_parent_id(self):
        return -2 if self.is_root else self.parent.get_id()

    def add_length(self, swc_node):
        self.path_length += swc_node.path_length
        self.xy_path_length += swc_node.xy_path_length
        self.z_path_length += swc_node.z_path_length

    def add_data(self, swc_node):
        self.path_length += swc_node.path_length
        self.xy_path_length += swc_node.xy_path_length
        self.z_path_length += swc_node.z_path_length
        self.volume += swc_node.volume
        self.surface_area += swc_node.surface_area

    def is_virtual(self):
        """Returns True iff the node is virtual.
        """
        return self._id < 0

    def is_regular(self):
        """Returns True iff the node is NOT virtual.
        """
        return self._id >= 0

    def distance(self, tn=None, mode=_3D):
        """ Returns the distance to another node.
        It returns 0 if either of the nodes is not regular.

        Args:
          tn : the target node for distance measurement
        """
        # make sure itself is a regular node
        if not self.is_regular():
            return 0.0

        # make sure tn is a valid swc node
        if isinstance(tn, SwcNode) and tn.is_regular():
            if mode == _2D:
                return self.get_center().distance_to_point_2d(tn.get_center())
            return self.get_center().distance(tn.get_center())

        # euc node is also acceptable
        if isinstance(tn, EuclideanPoint):
            if mode == _2D:
                return self.get_center().distance_to_point_2d(tn)
            return self.get_center().distance(tn)

        return 0.0

    def parent_distance(self):
        """ Returns the distance to it parent.
        """
        return self.distance(self.parent)

    def scale(self, sx, sy, sz, adjusting_radius=True):
        """Transform a node by scaling
        """

        self._pos[0] *= sx
        self._pos[1] *= sy
        self._pos[2] *= sz

        if adjusting_radius:
            self._radius *= math.sqrt(sx * sy)

    def is_isolated(self):
        if (self.parent is None or self.parent.get_id() == -1) and len(self.children) == 0:
            return True
        return False

    def to_swc_str(self, pid=None):
        if pid is not None:
            return "{} {} {} {} {} {} {}\n".format(
                self._id, self._type, self.get_x(), self.get_y(), self.get_z(), self._radius, pid
            )

        return "{} {} {} {} {} {} {}\n".format(
            self._id, self._type, self.get_x(), self.get_y(), self.get_z(), self._radius, self.parent.get_id()
        )

    def __str__(self):
        return "%d (%d): %s, %g" % (self._id, self._type, str(self.get_center()._pos), self._radius)


class SwcTree:
    """A class for representing one or more SWC trees.
    For simplicity, we always assume that the root is a virtual node.
    """

    def __init__(self):
        self._root = Make_Virtual()
        self._size = None
        self._total_length = None
        self._name = None

        self.id_set = set()
        self.depth_array = None
        self.LOG_NODE_NUM = None
        self.lca_parent = None
        self.node_list = None
        self.id_node_dict = None

    def is_comment(self, line):
        return line.strip().startswith("#")

    def root(self):
        return self._root

    def size(self):
        self._size = len(self.id_set)
        return self._size

    def regular_root(self):
        return self._root.children

    def _print(self):
        print (RenderTree(self._root).by_attr("_id"))

    def clear(self):
        self._root = Make_Virtual()

    def get_name(self):
        return self._name

    # warning: slow, don't use in loop
    def node_from_id(self, nid):
        niter = iterators.PreOrderIter(self._root)
        for tn in niter:
            if tn.get_id() == nid:
                return tn
        return None

    # warning: slow, don't use in loop
    def parent_id(self, nid):
        tn = self.node_from_id(nid)
        return tn.get_parent_id() if tn else None

    # warning: slow, don't use in loop
    def parent_node(self, nid):
        tn = self.node_from_id(nid)
        return tn.parent if tn else None

    # warning: slow, don't use in loop
    def child_list(self, nid):
        tn = self.node_from_id(nid)
        return tn.children if tn else None

    def load_list(self, lines):
        self.clear()
        nodeDict = dict()
        for line in lines:
            if not self.is_comment(line):
                #                     print line
                data = list(map(float, line.split()))
                #                     print(data)
                if len(data) == 7:
                    nid = int(data[0])
                    ntype = int(data[1])
                    pos = EuclideanPoint(center=data[2:5])
                    radius = data[5]
                    parentId = data[6]
                    if nid in self.id_set:
                        raise Exception("[Error: SwcTree.load]Same id {}".format(nid))
                    self.id_set.add(nid)
                    tn = SwcNode(nid=nid, ntype=ntype, radius=radius, center=pos)
                    nodeDict[nid] = (tn, parentId)

        for _, value in nodeDict.items():
            tn = value[0]
            parentId = value[1]
            if parentId == -1:
                tn.parent = self._root
            else:
                parentNode = nodeDict.get(parentId)
                if parentNode:
                    tn.parent = parentNode[0]

        for node in self.get_node_list():
            if node.parent is None:
                continue
            if node.parent.get_id() == -1:
                node._depth = 0
            else:
                node._depth = node.parent._depth + 1
                node.root_length = node.parent.root_length + node.parent_distance()

    def load(self, path):
        self.clear()
        self._name = os.path.basename(path)
        with open(path, "r") as fp:
            lines = fp.readlines()
            self.load_list(lines)

    def has_regular_node(self):
        return len(self.regular_root()) > 0

    def scale(self, sx, sy, sz, adjusting_radius=True):
        niter = iterators.PreOrderIter(self._root)
        for tn in niter:
            tn.scale(sx, sy, sz, adjusting_radius)

    def rescale(self, scale):
        for node in self.get_node_list():
            node.set_z(node.get_z() * scale[2])
            node.set_y(node.get_y() * scale[1])
            node.set_x(node.get_x() * scale[0])

    def length(self, force_update=False):
        if self._total_length is not None and force_update == False:
            return self._total_length

        node_list = self.get_node_list()
        result = 0
        for tn in node_list:
            if tn.is_virtual() or tn.parent.is_virtual():
                continue
            result += tn.parent_distance()

        return result

    def radius(self, nid):

        return self.node_from_id(nid).radius()

    def get_depth_array(self, node_num):
        self.depth_array = [0] * (node_num + 10)
        node_list = self.get_node_list()
        for node in node_list:
            self.depth_array[node.get_id()] = node.depth()

    # initialize LCA data structure in swc_tree
    def get_lca_preprocess(self, node_num=-1):
        if node_num == -1:
            node_num = self.size()
        self.get_depth_array(node_num)
        self.LOG_NODE_NUM = math.ceil(math.log(node_num, 2)) + 1
        self.lca_parent = np.zeros(shape=(node_num + 10, self.LOG_NODE_NUM), dtype=int)
        tree_node_list = self.get_node_list()

        for node in tree_node_list:
            if node.is_virtual():
                continue
            self.lca_parent[node.get_id()][0] = node.parent.get_id()
        for k in range(self.LOG_NODE_NUM - 1):
            for v in range(1, node_num + 1):
                if self.lca_parent[v][k] < 0:
                    self.lca_parent[v][k + 1] = -1
                else:
                    self.lca_parent[v][k + 1] = self.lca_parent[int(self.lca_parent[v][k])][k]
        return True

    # input node id of two swc_node, calculate LCA
    def get_lca(self, u, v):
        lca_parent = self.lca_parent
        LOG_NODE_NUM = self.LOG_NODE_NUM
        depth_array = self.depth_array

        if depth_array[u] > depth_array[v]:
            u, v = v, u
        for k in range(LOG_NODE_NUM):
            if depth_array[v] - depth_array[u] >> k & 1:
                v = lca_parent[v][k]
        if u == v:
            return u
        for k in range(LOG_NODE_NUM - 1, -1, -1):
            if lca_parent[u][k] != lca_parent[v][k]:
                u = lca_parent[u][k]
                v = lca_parent[v][k]
        return lca_parent[u][0]

    def align_roots(self, gold_tree, matches, DEBUG=False):
        offset = EuclideanPoint()
        stack = queue.LifoQueue()
        swc_test_list = self.get_node_list()

        for root in gold_tree.root().children:
            gold_anchor = np.array(root._pos)
            if root in matches.keys():
                test_anchor = np.array(matches[root]._pos)
            else:
                nearby_nodes = get_nearby_swc_node_list(
                    gold_node=root, test_swc_list=swc_test_list, threshold=root.radius() / 2
                )
                if len(nearby_nodes) == 0:
                    continue
                test_anchor = nearby_nodes[0]._pos

            offset._pos = (test_anchor - gold_anchor).tolist()
            if DEBUG:
                print ("off_set:x = {}, y = {}, z = {}".format(offset._pos[0], offset._pos[1], offset._pos[2]))

            stack.put(root)
            while not stack.empty():
                node = stack.get()
                if node.is_virtual():
                    continue

                node._pos[0] += offset._pos[0]
                node._pos[1] += offset._pos[1]
                node._pos[2] += offset._pos[2]

                for son in node.children:
                    stack.put(son)

    def change_root(self, new_root_id):
        stack = queue.LifoQueue()
        swc_list = self.get_node_list()
        list_size = max(self.id_set)
        vis_list = np.zeros(shape=(list_size + 10))
        pa_list = [None] * (list_size + 10)

        for node in swc_list:
            pa_list[node.get_id()] = node.parent
        new_root = self.node_from_id(new_root_id)

        stack.put(new_root)
        pa_list[new_root_id] = self.root()
        while not stack.empty():
            cur_node = stack.get()
            vis_list[cur_node.get_id()] = True
            for son in cur_node.children:
                if not vis_list[son.get_id()]:
                    stack.put(son)
                    pa_list[son.get_id()] = cur_node
            if (
                cur_node.parent is not None
                and cur_node.parent.get_id() != -1
                and not vis_list[cur_node.parent.get_id()]
            ):
                stack.put(cur_node.parent)
                pa_list[cur_node.parent.get_id()] = cur_node

        for i in range(1, len(swc_list)):
            swc_list[i].parent = None
        for i in range(1, len(swc_list)):
            swc_list[i].parent = pa_list[swc_list[i].get_id()]

    def type_clear(self, col=0, rt_color=2):
        node_list = self.get_node_list()
        for node in node_list:
            node._type = col
        for root in self.root().children:
            root._type = rt_color

    def radius_limit(self, x):
        node_list = self.get_node_list()
        for node in node_list:
            node._radius /= x

    def next_id(self):
        return max(self.id_set) + 1

    # use this function if son is a new node
    def add_child(self, swc_node, swc_son_node):
        if not isinstance(swc_son_node, SwcNode) or not isinstance(swc_node, SwcNode):
            return False
        nid = self.next_id()

        swc_son_node.set_id(nid)
        swc_son_node.parent = swc_node

        self.id_set.add(nid)
        return True

    # use this function if son use to be a part of this tree
    def link_child(self, pa, son):
        if not isinstance(pa, SwcNode) or not isinstance(son, SwcNode):
            return False
        son.parent = pa
        return True

    def remove_node(self, node):
        if not isinstance(node, SwcNode):
            return False
        pa = node.parent
        children = list(pa.children)
        children.remove(node)
        pa.children = tuple(children)

        for son in node.children:
            son.parent = self._root
        self.id_set.remove(node.get_id())
        return True

    def unlink_child(self, node):
        if not isinstance(node, SwcNode):
            return False
        pa = node.parent
        children = list(pa.children)
        children.remove(node)
        pa.children = tuple(children)
        return True

    def get_node_list(self, update=False):
        if self.node_list is None or update:
            self.node_list = []
            q = queue.LifoQueue()
            q.put(self._root)
            while not q.empty():
                cur = q.get()
                self.node_list.append(cur)
                for child in cur.children:
                    q.put(child)

        return self.node_list

    def sort_node_list(self, key="default"):
        """
        index:
            default: order by pre order
            id: order by id
        """
        if key == "default":
            self.get_node_list(update=True)
        if key == "id":
            self.node_list.sort(key=lambda node: node.get_id())

    def to_str_list(self):
        swc_node_list = self.get_node_list()
        swc_str = []
        for node in swc_node_list:
            if node.is_virtual():
                continue
            swc_str.append(node.to_swc_str())
        return "".join(swc_str)

    def get_copy(self):
        new_tree = SwcTree()
        swc_str = self.to_str_list().split("\n")
        new_tree.load_list(swc_str)
        return new_tree

    def get_id_node_dict(self):
        if self.id_node_dict is not None:
            return self.id_node_dict
        self.id_node_dict = {}
        for node in self.get_node_list():
            self.id_node_dict[node.get_id()] = node
        return self.id_node_dict

    def get_branch_swc_list(self):
        """
        get branch nodes (link to more than 3 nodes) of swc tree
        """
        swc_list = self.get_node_list()
        branch_list = []
        for node in swc_list:
            if node.is_virtual():
                continue
            if node.parent.is_virtual():
                if len(node.children) > 2:
                    branch_list.append(node)
            elif len(node.children) > 1:
                branch_list.append(node)
        return branch_list

    def get_leaf_swc_list(self):
        swc_list = self.get_node_list()
        leaf_list = []
        for node in swc_list:
            if node.is_virtual():
                continue
            if node.parent.is_virtual():
                if len(node.children) == 1:
                    leaf_list.append(node)
            if len(node.children) == 0:
                leaf_list.append(node)
        return leaf_list

    def set_node_type_by_topo(self, root_id=1):
        """
        root_id decide other nodes' id
        branch = root_id + 1
        leaf = root_id + 3
        normal node = root_id + 2
        """
        swc_list = self.get_node_list()
        for node in swc_list:
            if node.is_virtual():
                continue
            if node.parent.get_id() == -1:
                node._type = root_id
            elif len(node.children) > 1:
                node._type = root_id + 1
            elif len(node.children) == 1:
                node._type = root_id + 2
            else:
                node._type = root_id + 3


if __name__ == "__main__":
    import os
    from pyneval.io.swc_writer import swc_save

    path = "..\..\data\\test_data\\topo_metric_data"
    for file in os.listdir(path):
        tree = SwcTree()
        tree.load(os.path.join(path, file))
        if file[:4] == "gold":
            tree.set_node_type_by_topo(root_id=1)
        else:
            tree.set_node_type_by_topo(root_id=5)
        swc_save(tree, out_path=os.path.join(path, "s" + file))



import sys
import jsonschema

from pyneval.model import swc_node
from pyneval.metric.utils import km_utils
from pyneval.metric.utils import point_match_utils
from pyneval.metric.utils.metric_manager import get_metric_manager
from pyneval.metric.utils import bigraph_maxmatch_utils
metric_manager = get_metric_manager()


class CriticalNodeMetric(object):
    """
    critical node metric
    """
    def __init__(self, config):
        # read configs
        self.debug = config["debug"] if config.get("debug") is not None else False
        self.distance_threshold = config["distance_threshold"]
        self.critical_type = config["critical_type"]
        self.scale = config["scale"]
        self.true_positive = config["true_positive"]
        self.missed = config["missed"]
        self.excess = config["excess"]

    def get_result(self, test_len, gold_len, switch, km, threshold_dis):
        false_pos_num, false_neg_num, true_pos_num = 0, 0, 0
        # count number of nodes which are matched, calculate FP, TN, TP
        for i in range(gold_len):
            if km.match[i] != -1 and km.G[km.match[i]][i] != -0x3F3F3F3F / 2:
                true_pos_num += 1
        false_neg_num = gold_len - true_pos_num
        false_pos_num = test_len - true_pos_num

        # definition of swich is in function "get_dis_graph"
        if switch:
            false_neg_num, false_pos_num = false_pos_num, false_neg_num
            gold_len, test_len = test_len, gold_len

        if true_pos_num != 0:
            mean_dis = -km.get_max_dis() / true_pos_num
        else:
            mean_dis = 0.0
        if mean_dis == -0.0:
            mean_dis = 0.0

        pt_cost = -km.get_max_dis() + threshold_dis * (false_neg_num + false_pos_num) / (
            false_neg_num + false_pos_num + true_pos_num
        )

        # debug:
        # print("output")
        # print(false_pos_num)
        # print(true_neg_num)
        # print(mean_dis)
        # print(pt_cost)
        return gold_len, test_len, true_pos_num, false_neg_num, false_pos_num, mean_dis, mean_dis * true_pos_num, pt_cost

    def get_colored_tree(self, gold_node_list, test_node_list, MaxMatch, color):
        """
        color[0] = tp's color
        color[1] = fp's color
        color[2] = fn's color
        """
        for i in range(len(gold_node_list)):
            if MaxMatch.pa[i] != -1:
                gold_node_list[i]._type = color[0]
            else:
                gold_node_list[i]._type = color[2]
        for i in range(len(test_node_list)):
            if MaxMatch.pb[i] != -1:
                test_node_list[i]._type = color[0]
            else:
                test_node_list[i]._type = color[1]

    def KM_max_match(
            self,
        gold_tree,
        test_tree,
        test_node_list,
        gold_node_list,
        threshold_dis,
    ):
        """
        get minimum matching distance by running KM algorithm
        than calculate the return value according to matching result
        Args:
            gold_tree(Swc Tree)
            test_tree(Swc Tree)
            gold_node_list(List): contains only branch nodes
            test_node_list(List): contains only branch nodes
            threshold_dis: if the distance of two node are larger than this threshold,
                           they are considered unlimited far
        Returns:
            gold_len(int): length of gold_node_list
            test_len(int): length of test_node_list
            true_pos_num(int): number of nodes in both gold and test tree
            false_neg_num(int): number of nodes in gold but not test tree
            false_pos_num(int): number of nodes in test but not gold tree
            mean_dis: mean distance of nodes that are successfully matched(true positive)
            tot_dis: total distance of nodes that are successfully matched(true positive)
            pt_cost: a composite value calculated by tp, fn, fp and threshold
            iso_node_num: number of nodes in test tree without parents or children
        """
        test_gold_dict = point_match_utils.get_swc2swc_dicts(
            src_node_list=test_tree.get_node_list(), tar_node_list=gold_tree.get_node_list()
        )
        # disgraph is a 2D ndarray store the distance between nodes in gold and test
        # test_node_list contains only branch or leaf nodes
        dis_graph, switch, test_len, gold_len = km_utils.get_dis_graph(
            gold_tree=gold_tree,
            test_tree=test_tree,
            test_node_list=test_node_list,
            gold_node_list=gold_node_list,
            test_gold_dict=test_gold_dict,
            threshold_dis=threshold_dis,
            metric_mode=1,
        )
        # create a KM object and calculate the minimum match
        km = km_utils.KM(maxn=max(test_len, gold_len) + 10, nx=test_len, ny=gold_len, G=dis_graph)
        km.solve()
        # calculate the result
        gold_len, test_len, true_pos_num, false_neg_num, false_pos_num, mean_dis, tot_dis, pt_cost = self.get_result(
            test_len=test_len, gold_len=gold_len, switch=switch, km=km, threshold_dis=threshold_dis
        )
        # calculate the number of isolated nodes
        iso_node_num = 0
        for node in test_tree.get_node_list():
            if node.is_isolated():
                iso_node_num += 1
        return gold_len, test_len, true_pos_num, false_neg_num, false_pos_num, mean_dis, tot_dis, pt_cost, iso_node_num

    def max_match(self, test_node_list, gold_node_list, distance_threshold):
        gold_list_len = len(gold_node_list)
        test_list_len = len(test_node_list)
        MaxMatch = bigraph_maxmatch_utils.AugmentPath(gold_list_len, test_list_len)
        for i in range(len(gold_node_list)):
            for j in range(len(test_node_list)):
                dis = gold_node_list[i].distance(test_node_list[j])
                if dis <= distance_threshold:
                    MaxMatch.add(i, j)
        MaxMatch.solve()
        return MaxMatch

    def get_total_dis(self, gold_critical_swc_list, test_critical_swc_list, MaxMatch):
        dis = 0
        for i in range(len(MaxMatch.pa)):
            if MaxMatch.pa[i] == -1:
                continue
            dis += gold_critical_swc_list[i].distance(test_critical_swc_list[MaxMatch.pa[i]])
        return dis

    def run(self, gold_swc_tree, test_swc_tree):
        # rescale the input tree
        gold_swc_tree.rescale(self.scale)
        test_swc_tree.rescale(self.scale)

        # denote the color id of different type of nodes.
        color = [self.true_positive, self.missed, self.excess]
        gold_swc_tree.type_clear(0, 0)
        test_swc_tree.type_clear(0, 0)

        test_critical_swc_list = []
        gold_critical_swc_list = []
        # critical_type == 1: both branches and leaves
        # critical_type == 2: only leaves
        # critical_type == 3: only branches
        if self.critical_type != 2:
            test_critical_swc_list.extend(test_swc_tree.get_branch_swc_list())
            gold_critical_swc_list.extend(gold_swc_tree.get_branch_swc_list())
        if self.critical_type != 3:
            test_critical_swc_list.extend(test_swc_tree.get_leaf_swc_list())
            gold_critical_swc_list.extend(gold_swc_tree.get_leaf_swc_list())

        MaxMatch = self.max_match(
            test_node_list=test_critical_swc_list,
            gold_node_list=gold_critical_swc_list,
            distance_threshold=self.distance_threshold
        )
        # get a colored tree with
        self.get_colored_tree(
            gold_node_list=gold_critical_swc_list, test_node_list=test_critical_swc_list,
            MaxMatch=MaxMatch, color=color
        )

        precision = MaxMatch.res / len(test_critical_swc_list)
        recall = MaxMatch.res / len(gold_critical_swc_list)
        tot_dis = self.get_total_dis(gold_critical_swc_list, test_critical_swc_list, MaxMatch)
        branch_result = {
            "gold_len": len(gold_critical_swc_list),
            "test_len": len(test_critical_swc_list),
            "true_pos_num": MaxMatch.res,
            "false_neg_num": len(gold_critical_swc_list) - MaxMatch.res,
            "false_pos_num": len(test_critical_swc_list) - MaxMatch.res,
            "mean_dis": tot_dis / MaxMatch.res,
            "precision": precision,
            "recall": recall,
            "tot_dis": tot_dis,
            "f1_score": 2*precision*recall/(precision + recall),
        }
        return branch_result, gold_swc_tree, test_swc_tree

@metric_manager.register(
    name="cn",
    config="critical_node_metric.json",
    desc="quality of critical points",
    public=True,
    alias=["CN"],
)
def critical_node_metric(gold_swc_tree, test_swc_tree, config):
    """
    branch metric calculates the minimum distance match between branches of two swc trees
    This function is used for unpacking configs and packaging return values
    Args:
        gold_swc_tree(Swc Tree) gold standard tree
        test_swc_tree(Swc Tree) reconstructed tree
        config(dict):
            keys: the name of configs
            items: config values
    Example:
        test_tree = swc_node.SwcTree()
        gold_tree = swc_node.SwcTree()
        gold_tree.load("..\\..\\data\\test_data\\topo_metric_data\\gold_fake_data1.swc")
        test_tree.load("..\\..\\data\\test_data\\topo_metric_data\\test_fake_data1.swc")
        branch_result = length_metric(gold_swc_tree=gold_tree,
                                      test_swc_tree=test_tree,
                                      config=config)
        print(branch_result["mean_dis"])
        ...
    Return:
         branch_result(tuple) a tuple of 9 metric results
    """

    critical_node = CriticalNodeMetric(config)
    return critical_node.run(gold_swc_tree, test_swc_tree)



from anytree import NodeMixin
from pyneval.model.euclidean_point import EuclideanPoint
from pyneval.model.binary_node import BinaryNode

import copy
import math
import queue

_3D = "3d"
_2D = "2d"
TRAJECTORY_NONE = -1.0
FLOAT_INF = 999999999.9
RIGHT = 'right'
LEFT = 'left'
DEFULT = 'DEFULT'


# recurrently convert a swcnode tree into a binary tree
# input: root of a swcnode tree
# return: root of a binary tree
def swctree_to_binarytree(node):
    binary_root = BinaryNode(data=node)

    # the nodes in this list is the root of a binary tree
    binnary_son_list = []
    son = list(node.children)

    for son_node in son:
        # print(type(son_node))
        binnary_node = swctree_to_binarytree(son_node)
        binnary_son_list.append(binnary_node)

    while len(binnary_son_list) > 2:
        best1 = binnary_son_list[0]
        best2 = binnary_son_list[1]
        distance = FLOAT_INF
        for i in range(len(binnary_son_list)- 1):
            bin_node1 = binnary_son_list[i]
            for j in range(i+1, len(binnary_son_list)):
                bin_node2 = binnary_son_list[j]
                if bin_node1.data.distance(bin_node2.data) < distance:
                    best1 = bin_node1
                    best2 = bin_node2
                    distance = bin_node1.data.distance(bin_node2.data)

        # new_swcnode = copy.copy(node)
        new_binnode = BinaryNode(data=node, left_son=best1, right_son=best2)
        binnary_son_list.remove(best1)
        binnary_son_list.remove(best2)
        binnary_son_list.append(new_binnode)

    if len(binnary_son_list) >= 1:
        binary_root.left_son = binnary_son_list[0]
    if len(binnary_son_list) == 2:
        binary_root.right_son = binnary_son_list[1]

    return binary_root


def re_arrange(bin_node, hight=1, parent=None, side=DEFULT):
    bin_node.hight = hight
    bin_node.parent = parent
    bin_node.side = side
    bin_node.max_dep = 1
    bin_node.treesize = 1
    if bin_node.left_son is not None:
        re_arrange(bin_node=bin_node.left_son, hight=hight+1, parent=bin_node, side=LEFT)
        bin_node.max_dep = max(bin_node.max_dep, bin_node.left_son.max_dep + 1)
        bin_node.treesize += bin_node.left_son.treesize

    if bin_node.right_son is not None:
        re_arrange(bin_node=bin_node.right_son, hight=hight+1, parent=bin_node,side=RIGHT)
        bin_node.max_dep = max(bin_node.max_dep, bin_node.right_son.max_dep + 1)
        bin_node.treesize += bin_node.right_son.treesize


def calculate_trajectories_xy(origin, first, second, threshold):
    point = EuclideanPoint(center=[0,0,0])

    to_first = origin.distance(first, _2D)
    to_second = origin.distance(second, _2D)

    proportionAlongLine = (threshold - to_first)/(to_second - to_first)

    point.set_x(first.get_x() + proportionAlongLine*(second.get_x() - first.get_x()))
    point.set_y(first.get_y() + proportionAlongLine*(second.get_y() - first.get_y()))
    return point


def calculate_trajectories_z(origin, first, second, threshold):
    to_first = math.fabs(origin.get_z() - first.get_z())
    to_second = math.fabs(origin.get_z() - second.get_z())

    proportionAlongLine = (threshold - to_first)/(to_second - to_first)

    return first.get_z() + proportionAlongLine*(second.get_z() - first.get_z())


def find_parent_trajectory(node, thereholds, DEBUG=False):
    done_x = False
    done_z = False
    xy_dis = 0.0
    z_dis = 0.0

    c = node
    data = node.data
    c_data = c.data

    point = EuclideanPoint(center=[0,0,0])
    while not done_x or not done_z:
        c = c.parent
        prev_data = c_data
        c_data = c.data
        if not done_x:
            xy_dis = data.distance(c_data, _2D)
            if xy_dis > thereholds.get_x():
                tmp_point = EuclideanPoint(center=[0, 0, 0])
                tmp_point.add_coord(calculate_trajectories_xy(data, prev_data, c_data, thereholds.get_x()))
                point.set_x(tmp_point.get_x())
                point.set_y(tmp_point.get_y())
                done_x = True
            elif c.is_root:
                point.set_x(c_data.get_x())
                point.set_y(c_data.get_y())
                done_x = True

        if not done_z:
            z_dis = math.fabs(data.get_z() - c_data.get_z())
            if z_dis > thereholds.get_z():
                point.set_z(calculate_trajectories_z(data, prev_data, c_data, thereholds.get_z()))
                done_z = True
            elif c.is_root:
                point.set_z(c_data.get_z())
                done_z = True
    if DEBUG:
        print(node.data._id)
        print("trx = {}, try = {}, trz = {}".format(point.get_x(), point.get_y(), point.get_z()))
        print("----------------------------------------------------------------")
    return point


def find_child_trajectory(data, child, thresholds, DEBUG=False):
    xy_dis = 0.0
    z_dis = 0.0
    done_x = False
    done_z = False
    prev_data = data

    point = EuclideanPoint(center=[0, 0, 0])
    while not done_x or not done_z:
        c_data = child.data

        if not done_x:
            xy_dis = data.distance(c_data, _2D)

            if xy_dis > thresholds.get_x():
                tmp_point = EuclideanPoint(center=[0, 0, 0])
                tmp_point.add_coord(calculate_trajectories_xy(data, prev_data, c_data, thresholds.get_x()))
                point.set_x(tmp_point.get_x())
                point.set_y(tmp_point.get_y())
                done_x = True
            elif child.is_leaf():
                point.set_x(c_data.get_x())
                point.set_y(c_data.get_y())
                done_x = True
            elif child.right_son != None:
                point.set_x(TRAJECTORY_NONE)
                point.set_y(TRAJECTORY_NONE)
                done_x = True

        if not done_z:
            z_dis = math.fabs(data.get_z() - c_data.get_z())
            if z_dis > thresholds.get_z():
                point.set_z(calculate_trajectories_z(data, prev_data, c_data, thresholds.get_z()))
                done_z = True
            elif child.is_leaf():
                point.set_z(c_data.get_z())
                done_z = True
            elif child.right_son != None:
                point.set_z(TRAJECTORY_NONE)
                done_z = True

        prev_data = c_data
        if not child.is_leaf():
            child = child.left_son

    if DEBUG:
        print(data.id)
        print("trx = {}, try = {}, trz = {}".format(point.get_x(), point.get_y(), point.get_z()))
        print("----------------------------------------------------------------")
    return point


def add_child_bifurcations(stack, node):
    lson = node.left_son
    rson = node.right_son

    while lson.left_son is not None and lson.right_son is None:
        lson = lson.left_son
    stack.put(lson)

    while rson.left_son is not None and rson.right_son is None:
        rson = rson.left_son
    stack.put(rson)


# calculate the distance to root
def calculate_trajectories(swc_root, bin_root, thresholds, z_in_path_dist =True, current_trajectories=0.0, DEBUG=False):
    stack = queue.LifoQueue()

    node = bin_root
    while node.left_son != None and node.right_son == None:
        node = node.left_son
    stack.put(node)

    while not stack.empty():
        node = stack.get()
        if DEBUG:
            print("[debug:  ] calculate trajectories for node {}".format(node.data.get_id()))
        data = node.data

        if data.parent_trajectory is None:
            data.parent_trajectory = EuclideanPoint(center=[0, 0, 0])
        if not node.is_root():
            data.parent_trajectory.add_coord(find_parent_trajectory(node, thresholds))
        else:
            tmp_pa = BinaryNode(data=swc_root)
            node.parent = tmp_pa
            data.parent_trajectory.add_coord(find_parent_trajectory(node, thresholds))
            node.parent = None

        if not node.is_leaf():
            if data.left_trajectory is None:
                data.left_trajectory = EuclideanPoint(center=[0, 0, 0])
            if data.right_trajectory is None:
                data.right_trajectory = EuclideanPoint(center=[0, 0, 0])
            data.left_trajectory.add_coord(find_child_trajectory(data, node.left_son, thresholds))
            data.right_trajectory.add_coord(find_child_trajectory(data, node.right_son, thresholds))
            add_child_bifurcations(stack, node)

        if DEBUG:
            print("trajectory:")
            if data.parent_trajectory is not None:
                print("pa_tra = {}".format(data.parent_trajectory._pos))
            if data.left_trajectory is not None:
                print("left_tra = {}".format(data.left_trajectory._pos))
            if data.right_trajectory is not None:
                print("right_tra = {}".format(data.right_trajectory._pos))


def calculate_path_data(bin_node, z_in_path_dist, DEBUG=False):
    data = bin_node.data
    pa_data = bin_node.parent.data
    if pa_data.get_id() == -1:
        return

    if data.path_length == 0.0:
        if z_in_path_dist:
            data.path_length = data.distance(pa_data)
            data.xy_path_length = data.distance(pa_data, mode=_2D)
        else:
            data.path_length = data.distance(pa_data, mode=_2D)
            data.xy_path_length = data.distance(pa_data, mode=_2D)
        data.z_path_length = math.fabs(pa_data.get_z() - data.get_z())
        data.surface_area = data.path_length*math.pi*data.radius()*2
        data.volume = data.path_length*math.pi*data.radius()*data.radius()
    if DEBUG:
        print("node_id = {}, path = {}, xy_path = {}, z_path = {}".format(data._id,
                                                                          data.path_length,
                                                                          data.xy_path_length,
                                                                          data.z_path_length))


def remove_continuations(swc_root, bin_root, calc_path_dist, z_in_path_dist):
    stack = queue.LifoQueue()

    # swc_node: data
    child_data = None

    # bin_node: node
    res_root = bin_root
    child = None

    if calc_path_dist:
        data = bin_root.data

        if z_in_path_dist:
            data.path_length = swc_root.distance(data)
            data.xy_path_length = swc_root.distance(data, mode="2d")
        else:
            data.path_length = swc_root.distance(data, mode="2d")
            data.xy_path_length = swc_root.distance(data, mode="2d")
        data.z_path_length = math.fabs(swc_root.get_z() - bin_root.data.get_z())

    stack.put(bin_root)
    while not stack.empty():
        node = stack.get()
        data = node.data

        if not node.is_leaf():
            stack.put(node.left_son)
            if calc_path_dist:
                calculate_path_data(node.left_son, z_in_path_dist)
            if node.right_son is None:
                # prepare to remove the node
                child = node.left_son
                child_data = child.data
                if calc_path_dist:
                    child_data.add_data(data)
                if node == res_root:
                    res_root = child
                    child.parent = None
                else:
                    if node.get_side() == LEFT:
                        node.parent.left_son = child
                    elif node.get_side() == RIGHT:
                        node.parent.right_son = child
                    child.parent = node.parent
            else:
                stack.put(node.right_son)
                if calc_path_dist:
                    calculate_path_data(node.right_son, z_in_path_dist)
    return res_root


def convert_to_binarytree(tot_root, swc_root):
    # swc_tree._print()
    bintree_root = swctree_to_binarytree(swc_root)
    # debug
    # bintree_root.print_tree()

    re_arrange(bintree_root)
    calculate_trajectories(swc_root=tot_root,
                           bin_root=bintree_root,
                           thresholds=EuclideanPoint([1.2,0,0]),
                           z_in_path_dist=True)

    bintree_root = remove_continuations(swc_root=tot_root,
                                        bin_root=bintree_root,
                                        calc_path_dist=True,
                                        z_in_path_dist=False)
    return bintree_root


def convert_to_binarytrees(root):
    tot_root = BinaryNode(data=root)
    bin_root_list = []
    for node in root.children:
        bin_node = convert_to_binarytree(root, node)
        bin_root_list.append(bin_node)
    return tot_root, bin_root_list



import numpy

import os
import urllib
import gzip
import cPickle as pickle

def mnist_generator(data, batch_size, n_labelled, limit=None):
    images, targets = data

    rng_state = numpy.random.get_state()
    numpy.random.shuffle(images)
    numpy.random.set_state(rng_state)
    numpy.random.shuffle(targets)
    if limit is not None:
        print "WARNING ONLY FIRST {} MNIST DIGITS".format(limit)
        images = images.astype('float32')[:limit]
        targets = targets.astype('int32')[:limit]
    if n_labelled is not None:
        labelled = numpy.zeros(len(images), dtype='int32')
        labelled[:n_labelled] = 1

    def get_epoch():
        rng_state = numpy.random.get_state()
        numpy.random.shuffle(images)
        numpy.random.set_state(rng_state)
        numpy.random.shuffle(targets)

        if n_labelled is not None:
            numpy.random.set_state(rng_state)
            numpy.random.shuffle(labelled)

        image_batches = images.reshape(-1, batch_size, 784)
        target_batches = targets.reshape(-1, batch_size)

        if n_labelled is not None:
            labelled_batches = labelled.reshape(-1, batch_size)

            for i in range(len(image_batches)):
                yield (numpy.copy(image_batches[i]), numpy.copy(target_batches[i]), numpy.copy(labelled))

        else:

            for i in range(len(image_batches)):
                yield (numpy.copy(image_batches[i]), numpy.copy(target_batches[i]))

    return get_epoch

def load(batch_size, test_batch_size, n_labelled=None):
    filepath = '/tmp/mnist.pkl.gz'
    url = 'http://www.iro.umontreal.ca/~lisa/deep/data/mnist/mnist.pkl.gz'

    if not os.path.isfile(filepath):
        print "Couldn't find MNIST dataset in /tmp, downloading..."
        urllib.urlretrieve(url, filepath)

    with gzip.open('/tmp/mnist.pkl.gz', 'rb') as f:
        train_data, dev_data, test_data = pickle.load(f)

    return (
        mnist_generator(train_data, batch_size, n_labelled), 
        mnist_generator(dev_data, test_batch_size, n_labelled), 
        mnist_generator(test_data, test_batch_size, n_labelled)
    )


DINF = 999999999999.0
EPS = 0.0000001
SAME_POS_TH = 0.03
MIN_SIZE = 0.8
FLOAT_ERROR = 0.001


def get_avg_radius(gold_swc_tree, rate=0.1):
    total_length = gold_swc_tree.length()
    total_node = gold_swc_tree.size()
    if total_node <= 1:
        dis_threshold = 0.1
    else:
        dis_threshold = (total_length/total_node)*rate
    return dis_threshold


def get_screen_output():
    res = {
        "ssd": ["ssd_score", "recall", "precision", "f1_score"],
        "length": ["recall", "precision", "f1_score"],
        "cn": ["mean_dis", "recall", "precision", "f1_score"],
        "diadem": ["diadem_score"],
    }
    return res


def get_default_configs(method):
    configs = dict()
    if method == "ssd":
        configs["threshold_mode"] = 1
        configs["ssd_threshold"] = 2
        configs["up_sample_threshold"] = 2
        configs["scale"] = [1, 1, 1]
        configs["debug"] = False

    if method == "length":
        configs["radius_mode"] = 1
        configs["radius_threshold"] = 2
        configs["length_threshold"] = 0.2
        configs["scale"] = [1, 1, 1]
        configs["debug"] = False

    if method == "diadem":
        configs["weight_mode"] = 1
        configs["remove_spur"] = False
        configs["count_excess_nodes"] = True
        configs["align_tree_by_root"] = False
        configs["list_miss"] = True
        configs["list_distant_matches"] = True
        configs["list_continuations"] = True
        configs["find_proper_root"] = True
        configs["scale"] = [1, 1, 1]
        configs["xy_threshold"] = 2
        configs["z_threshold"] = 1
        configs["default_xy_path_error_threshold"] = 0.05
        configs["default_z_path_error_threshold"] = 0.05
        configs["local_path_error_threshold"] = 0.4
        configs["debug"] = False

    if method == "cn":
        configs["distance_threshold"] = 2
        configs["critical_type"] = 1
        configs["scale"] = [1, 1, 1]
        configs["true_positive"] = 3
        configs["missed"] = 4
        configs["excess"] = 5

    if method == "link":
        configs["scale"] = [1, 1, 1]

    if method == "volume":
        configs["length_threshold"] = 1.0
        configs["intensity_threshold"] = 1.0
        configs["debug"] = False

    return configs


def get_config_schema(method):
    schema = dict()

    if method == "ssd":
        schema = {
            "title": "ssd metric schema",
            "type": "object",
            "required": [
                "threshold_mode",
                "ssd_threshold",
                "up_sample_threshold",
                "scale",
                "debug"
            ],
            "properties": {
                "threshold_mode": {"type": "integer", "enum": [1, 2]},
                "ssd_threshold": {"type": "number", "minimum": 0},
                "up_sample_threshold": {"type": "number", "minimum": 0},
                "scale": {
                    "type": "array",
                    "additionalItems": {"type": "number", "minimum": 0},
                    "minItems": 3,
                    "maxItems": 3
                },
                "debug": {"type": "boolean"}
            }
        }

    if method == "length":
        schema = {
            "title": "length metric schema",
            "type": "object",
            "required": [
                "radius_mode",
                "radius_threshold",
                "length_threshold",
                "scale",
                "debug"
            ],
            "properties": {
                "radius_mode": {"type": "number", "enum": [1,2]},
                "radius_threshold": {"type": "number", "minimum": 0},
                "length_threshold": {"type": "number", "minimum": 0},
                "scale": {
                    "type": "array",
                    "additionalItems": {"type": "number", "minimum": 0},
                    "minItems": 3,
                    "maxItems": 3
                },
                "debug": {"type": "boolean"}
            }
        }

    if method == "diadem":
        schema = {
            "title": "diadem metric schema",
            "type": "object",
            "required": [
                "weight_mode"
            ],
            "properties": {
                "weight_mode": {"type": "number", "enum": [1,2,3,4,5]},
                "remove_spur": {"type": "boolean", "default": "false"},
                "count_excess_nodes": {"type": "boolean", "default": "true"},
                "align_tree_by_root": {"type": "boolean", "default": "false"},
                "list_miss": {"type": "boolean", "default": "false"},
                "list_distant_matches": {"type": "boolean", "default": "false"},
                "list_continuations": {"type": "boolean", "default": "false"},
                "find_proper_root": {"type": "boolean", "default": "true"},
                "scale": {
                    "type":"array",
                    "additionalItems": {"type": "number", "minimum": 0},
                    "minItems": 3,
                    "maxItems": 3
                },
                "xy_threshold": {"type": "number", "minimum": 0},
                "z_threshold": {"type": "number", "minimum": 0},
                "default_xy_path_error_threshold": {"type": "number", "minimum": 0},
                "default_z_path_error_threshold": {"type": "number", "minimum": 0},
                "local_path_error_threshold": {"type": "number", "minimum": 0},
                "debug": {"type": "boolean", "default": "false"}
            }
        }

    if method == "cn":
        schema = {
            "title": "branch metric schema",
            "type": "object",
            "required": [
              "distance_threshold",
              "critical_type",
              "scale",
              "true_positive",
              "missed",
              "excess"
            ],
            "properties": {
                "threshold_dis": {"type": "number", "minimum": 0},
                "threshold_mode": {"type": "number", "enum": [1, 2]},
                "critical_type": {"type": "integer", "enum": [1, 2, 3]},
                "scale": {
                    "type": "array",
                    "additionalItems": {"type": "number", "minimum": 0},
                    "minItems": 3,
                    "maxItems": 3
                },
                "true_positive": {"type": "integer", "exclusiveMinimum": 0},
                "missed": {"type": "integer", "exclusiveMinimum": 0},
                "excess": {"type": "integer", "exclusiveMinimum": 0}
            }
        }

    if method == "link":
        schema = {
            "title": "link metric schema",
            "type": "object",
            "required": [
              "scale"
            ],
            "properties": {
                "scale": {
                    "type": "array",
                    "additionalItems": {"type": "number", "minimum": 0},
                    "minItems": 3,
                    "maxItems": 3
                }
            }
        }

    if method == "volume":
        schema = {
            "title": "length metric schema",
            "type": "object",
            "required": [
              "length_threshold",
              "intensity_threshold",
              "debug"
            ],
            "properties": {
                "length_threshold": {"type": "number", "minimum": 0},
                "intensity_threshold": {"type": "number", "minimum": 0},
                "debug": {"type": "boolean"}
            }
        }

    return schema



import math

from pyneval.errors.exceptions import InvalidEuclideanPoint
from pyneval.metric.utils.config_utils import EPS


class EuclideanPoint(object):
    """
    geometry node without volume
    for point-line calculate
    """

    def __init__(self, center=None):
        center = center if center is not None else [0, 0, 0]
        if not isinstance(center, list):
            raise Exception("[Error: ]not a list")
        self._pos = center

    def get_x(self):
        return self._pos[0]

    def get_y(self):
        return self._pos[1]

    def get_z(self):
        return self._pos[2]

    def set_x(self, x):
        self._pos[0] = x

    def set_y(self, y):
        self._pos[1] = y

    def set_z(self, z):
        self._pos[2] = z

    def to_str(self):
        print ("{} {} {}".format(self._pos[0], self._pos[1], self._pos[2]))

    def add_coord(self, point_a):
        """
        :param point_a: another EuclideanPoint
        :return: True if success
        """
        if not isinstance(point_a, EuclideanPoint):
            return False

        self._pos[0] += point_a.get_x()
        self._pos[1] += point_a.get_y()
        self._pos[2] += point_a.get_z()

        return True

    def get_foot_point(self, line):
        p = self._pos
        a = line.coords[0]
        b = line.coords[1]
        a_p = [a[0] - p[0], a[1] - p[1], a[2] - p[2]]
        b_a = [b[0] - a[0], b[1] - a[1], b[2] - a[2]]

        k_up = -(a_p[0] * b_a[0] + a_p[1] * b_a[1] + a_p[2] * b_a[2])
        k_down = b_a[0] ** 2 + b_a[1] ** 2 + b_a[2] ** 2
        if k_down < EPS:
            raise Exception("[Error: ] Line {} {} is just a point".format(a, b))
        k = k_up / k_down
        foot = [k * b_a[0] + a[0], k * b_a[1] + a[1], k * b_a[2] + a[2]]
        return EuclideanPoint(foot)

    def get_closest_point(self, line):
        foot = self.get_foot_point(line)
        if foot.on_line(line):
            return foot
        else:
            dis1 = self.distance(EuclideanPoint(center=line.coords[0]))
            dis2 = self.distance(EuclideanPoint(center=line.coords[1]))
            if dis1 < dis2:
                return EuclideanPoint(center=line.coords[0])
            else:
                return EuclideanPoint(center=line.coords[1])

    def on_line(self, line):
        p = self._pos
        a = line.coords[0]
        b = line.coords[1]

        if (
            min(a[0], b[0]) <= p[0] <= max(a[0], b[0])
            and min(a[1], b[1]) <= p[1] <= max(a[1], b[1])
            and min(a[2], b[2]) <= p[2] <= max(a[2], b[2])
        ):
            return True
        return False

    def distance_to_coord(self, coord):
        if not isinstance(coord, list) and len(coord) != 3:
            raise InvalidEuclideanPoint(coord)
        point = EuclideanPoint(coord)
        return self.distance_to_point(point)

    def distance_to_point(self, point):
        if not isinstance(point, EuclideanPoint):
            raise InvalidEuclideanPoint(point)
        sub = [self._pos[0] - point._pos[0], self._pos[1] - point._pos[1], self._pos[2] - point._pos[2]]
        return math.sqrt(sub[0] * sub[0] + sub[1] * sub[1] + sub[2] * sub[2])

    def distance_to_point_2d(self, point):
        if not isinstance(point, EuclideanPoint):
            raise InvalidEuclideanPoint(point)

        sub = [self._pos[0] - point._pos[0], self._pos[1] - point._pos[1], self._pos[2] - point._pos[2]]
        return math.sqrt(sub[0] * sub[0] + sub[1] * sub[1])

    def distance_to_line(self, line):
        foot = self.get_foot_point(line)
        return self.distance_to_point(foot)

    def distance_to_segment(self, line):
        foot = self.get_foot_point(line)
        # foot = EuclideanPoint([0, 0, 0])
        if foot.on_line(line):
            return self.distance_to_point(foot)
        else:
            dis1 = self.distance_to_coord(line.coords[0])
            dis2 = self.distance_to_coord(line.coords[1])
            return min(dis1, dis2)

    def distance(self, obj):
        if isinstance(obj, EuclideanPoint):
            return self.distance_to_point(obj)
        elif isinstance(obj, Line):
            if obj.is_segment:
                return self.distance_to_segment(obj)
            else:
                return self.distance_to_line(obj)
        else:
            raise Exception("[Error: ] unexpected object type {}".format(type(obj)))


class Line:
    """
    consist of two EuclideanPoint
    coords[0] and coords[1]
    """

    def __init__(self, coords=None, e_node_1=None, e_node_2=None, is_segment=True):
        if coords is not None:
            self.coords = coords
        else:
            self.coords = [[], []]
            self.coords[0] = e_node_1._pos
            self.coords[1] = e_node_2._pos
        self.is_segment = is_segment

    def to_str(self):
        print (
            "[{} {} {}, {} {} {}]".format(
                self.coords[0].get_x(),
                self.coords[0].get_y(),
                self.coords[0].get_z(),
                self.coords[1].get_x(),
                self.coords[1].get_y(),
                self.coords[1].get_z(),
            )
        )

    def get_points(self):
        point_a = EuclideanPoint(self.coords[0])
        point_b = EuclideanPoint(self.coords[1])
        return point_a, point_b

    def distance(self, obj):
        if isinstance(obj, EuclideanPoint):
            return obj.distance(self)
        else:
            raise Exception("[Error: ] unexpected object type" + type(obj))



import queue
import kdtree
from anytree import PreOrderIter
from pyneval.model.swc_node import SwcTree
from pyneval.metric.utils.config_utils import SAME_POS_TH


def create_kdtree(node_list):
    """
    construct a kd-tree using nodes in the node list.
    Args:
        node_list(List): a list of Swc Nodes
    Return:
        my_kdtree(KD tree)
        pos_node_dict(Dict):
            key: tuple of (x, y, z) coordinate
            value: Swc node corresponds to the key's position
    """
    pos_node_dict = {}
    center_list = []

    for node in node_list:
        if node.is_virtual():
            continue
        pos_node_dict[node.get_center_as_tuple()] = node
        center_list.append(list(node.get_center_as_tuple()))
    my_kdtree = kdtree.create(center_list)
    return my_kdtree, pos_node_dict


def get_swc2swc_dicts(src_node_list, tar_node_list):
    '''
    get a dict mapping nodes from the source node list to the target node list
    two nodes are mapped if they are on the exactly same position (dis < SAME_POS_TH)
    Args:
        src_node_list(List): a list of Swc Nodes. find a match for each node in this list if possible.
        tar_node_list(List): a list of Swc Nodes. for src node list to search for matching target.
    Return:
        src_tar_dict(Dict):
            keys: node in src
            value: node in tar
    '''
    target_kd, pos_node_dict = create_kdtree(tar_node_list)

    src_tar_dict = {}

    for src_node in src_node_list:
        if src_node.is_virtual():
            continue
        # find the closest pos for gold node
        target_pos = target_kd.search_knn(list(src_node.get_center_as_tuple()), k=1)[0]
        target_node = pos_node_dict[tuple(target_pos[0].data)]
        # only if gold and test nodes are very close(dis < 0.03), they can be considered as the same pos
        if src_node.distance(target_node) < SAME_POS_TH:
            src_tar_dict[src_node] = target_node
        else:
            src_tar_dict[src_node] = None

    return src_tar_dict



import matplotlib.pyplot as plt
import numpy as np

data = [5, 20, 15, 25, 10]


def bar_visual():
    size = 4
    x = np.arange(size)

    a = [0.8044, 0.8312, 0.7218, 0.7527]
    b = [0.8021, 0.8354, 0.7358, 0.7537]
    c = [0.7899, 0.8335, 0.7326, 0.7318]
    d = [0.8039, 0.8318, 0.7367, 0.7588]
    e = [0.7825, 0.8192, 0.6796, 0.7310]

    total_width, n = 0.6, 5
    width = total_width / n
    x = x - (total_width - width) / 2

    ax = plt.axes()

    plt.ylim(0.66, 0.86)
    plt.xticks([0, 1, 2, 3], ["FM3", "FM4", "FM5", "FM6"])
    plt.xlabel("Test data")
    plt.ylabel("F1 score of SSD metric")
    ax.spines['top'].set_visible(False)
    ax.spines['right'].set_visible(False)

    plt.bar(x, [0, 0, 0, 0], width=width, edgecolor="k", label='self optimization', color='white')

    plt.bar(x, a, width=width, label='cross optimization', edgecolor="k", color='white', hatch='...')
    plt.bar(x + 1 * width, b, width=width, edgecolor="k", color='white', hatch='...')
    plt.bar(x + 2 * width, c, width=width, edgecolor="k", color='white', hatch='...')
    plt.bar(x + 3 * width, d, width=width, edgecolor="k", color='white', hatch='...')
    plt.bar(x + 4 * width, e, width=width, label='default', edgecolor="k", color='white', hatch='//')

    plt.bar(x, [0.8044, 0, 0, 0], width=width, edgecolor="k", color='white')
    plt.bar(x + 1 * width, [0, 0.8354, 0, 0], width=width, edgecolor="k", color='white')
    plt.bar(x + 2 * width, [0, 0, 0.7326, 0], width=width, edgecolor="k", color='white')
    plt.bar(x + 3 * width, [0, 0, 0, 0.7588], width=width, edgecolor="k", color='white')

    plt.legend()
    plt.show()


if __name__ == "__main__":
    bar_visual() 



from setuptools import setup

setup(name='pyneval',
      version='1.1.0',
      description='pyneval first edition',
      author='ZhangHan',
      author_email='benniehan98@gmail.com',
      url='https://github.com/bennieHan/PyNeval',
      packages=['pyneval', 'pyneval.cli', 'pyneval.io', 'pyneval.metric',
                'pyneval.metric.utils', 'pyneval.metric.utils.klib', 'pyneval.model',
                'pyneval.tools', 'pyneval.tools.optimize', 'pyneval.errors'],
      py_modules=['pyneval', 'pyneval.cli', 'pyneval.io', 'pyneval.metric',
                  'pyneval.metric.utils', 'pyneval.metric.utils.klib', 'pyneval.model',
                  'pyneval.tools', 'pyneval.tools.optimize', 'pyneval.errors'],
      install_requires=[
            'anytree>=2.7.2',
            'kdtree>=0.16',
            'numpy>=1.0',
            'rtree>=0.8',
            'jsonschema>=3.2.0',
            'pandas>=1.3.0',
            'matplotlib>=3.0.0'
      ],
      entry_points={
          'console_scripts': [
              'pyneval=pyneval.cli.pyneval:run'
          ]
      }
)



import jsonschema
import queue
import math

from pyneval.metric.utils.klib.TiffFile import imread
from pyneval.model.swc_node import SwcTree
from pyneval.tools.re_sample import up_sample_swc_tree

from pyneval.metric.utils.metric_manager import get_metric_manager

metric_manager = get_metric_manager()

class VolumeMetric(object):
    """
    volume metric
    """

    def __init__(self, config):
        self.debug = config["debug"] if config.get("debug") is not None else False
        self.length_threshold = config['length_threshold']
        self.intensity_threshold = config['intensity_threshold']

    def adjust_swc_tree(self, swc_tree, tiff_file, thres_intensity, max_step=4):
        swc_list = swc_tree.get_node_list()
        for node in swc_list:
            if node.is_virtual():
                continue
            new_center = cal_label(node=node, test_tiff=tiff_file,
                                   thres_intensity=thres_intensity, max_step=max_step)
            if not new_center:
                print("{} {} {} ".format(
                    node.get_x(), node.get_y(), node.get_z()
                ))
                print("[Warning: ] step too little")
            else:
                node.set_x(new_center[0])
                node.set_y(new_center[1])
                node.set_z(new_center[2])

    @staticmethod
    def get_dis(tuple_1, tuple_2):
        dx = tuple_1[0] - tuple_2[0]
        dy = tuple_1[1] - tuple_2[1]
        dz = tuple_1[2] - tuple_2[2]
        return math.sqrt(dx**2 + dy**2 + dz**2)

    def cal_label(self, node, test_tiff, thres_intensity, max_step=0):
        que = queue.Queue()
        center = [round(node.get_x()), round(node.get_y()), round(node.get_z())]
        que.put(tuple([center, 0]))
        vis = set()
        vis.add(tuple(center))

        stp = [
            [1, 0, 0], [0, 1, 0], [0, 0, 1],
            [-1, 0, 0], [0, -1, 0], [0, 0, -1]
        ]

        while not que.empty():
            cur = que.get()
            cur_loc = cur[0]
            cur_step = cur[1]
            if cur_step > max_step:
                continue

            if 0 <= round(cur_loc[0]) < test_tiff.shape[2] and\
               0 <= round(cur_loc[1]) < test_tiff.shape[1] and\
               0 <= round(cur_loc[2]) < test_tiff.shape[0] and\
               test_tiff[round(cur_loc[2])][round(cur_loc[1])][round(cur_loc[0])] >= thres_intensity:
                return tuple([round(cur_loc[0]), round(cur_loc[1]), round(cur_loc[2])])

            for i in range(6):
                dx = cur_loc[0] + stp[i][0]
                dy = cur_loc[1] + stp[i][1]
                dz = cur_loc[2] + stp[i][2]
                new_pos = tuple([dx, dy, dz])
                if new_pos not in vis:
                    vis.add(tuple([new_pos, cur_step + 1]))
                    que.put(tuple([new_pos, cur_step + 1]))
        return None


    def cal_volume_recall(self, test_tiff, gold_swc, intensity_threshold):
        swc_node_list = gold_swc.get_node_list()
        tot_front, tot_back = 0, 0

        for node in swc_node_list:
            if node.is_virtual():
                continue
            if (len(node.children) == 0 or node.parent.is_virtual()) and\
                    self.cal_label(node, test_tiff, intensity_threshold, 0) is not None:
                tot_front += 1
            elif self.cal_label(node, test_tiff, intensity_threshold, 1) is not None:
                tot_front += 1
            else:
                if self.debug:
                    print("[Info: ] fail: {} {} {} {} {}".format(
                        node.get_id(), round(node.get_x()), round(node.get_y()), round(node.get_z()),
                        test_tiff[round(node.get_z())][round(node.get_y())][round(node.get_x())]
                    ))

            tot_back += 1
        return tot_front/tot_back

    def run(self, gold_swc_tree, test_swc_tree):

        densed_swc_tree = up_sample_swc_tree(swc_tree=gold_swc_tree, length_threshold=self.length_threshold)
        recall = self.cal_volume_recall(test_swc_tree, densed_swc_tree, self.intensity_threshold)

        res = {
            "recall": recall
        }
        return res, None, None


@metric_manager.register(
    name="volume",
    config="volume_metric.json",
    desc="volume overlap",
    public=False,
    alias=['VM']
)
def volume_metric(gold_swc_tree, test_swc_tree, config):
    volume = VolumeMetric(config)
    return volume.run(gold_swc_tree, test_swc_tree)



from .glib.DrawSimulationSWCModel import simulate3DTreeModel
from .glib.ImageUtils import gaussianNoisyAddGray3D


def simulator(size,max_depth,base_length,base_x,mean,std):
	img,_,_=simulate3DTreeModel(size,max_depth,base_length,base_x)
	return gaussianNoisyAddGray3D(img,mean,std)


if __name__=='__main__':
	img=simulator(200,5,50,1,30,10)
	print(img)


# @Author: guanwanxian
# @Date:   1970-01-01T08:00:00+08:00
# @Email:  guanwanxian@zju.edu.cn
# @Last modified by:   guanwanxian
# @Last modified time: 2017-03-26T19:39:24+08:00

from .Transformations import superimposition_matrix
import numpy as np

class Point3D():
    """ 3D Object Class
    """
    def __init__(self, x, y, z):
        self.x = x
        self.y = y
        self.z = z

    def __str__(self):
        return 'Point3D: position {}'.format((self.x, self.y, self.z))

    def toList(self):
        return [self.x, self.y, self.z]

    def medianWithPoint(self, p):
        '''medianWithPoint: Calculate median with other Point3D
        Args:
            p (Point3D): The other point.

        Returns:
            (Point3D): The median point.
        '''
        x_m = self.x + (p.x-self.x)/2;
        y_m = self.y + (p.y-self.y)/2;
        z_m = self.z + (p.z-self.z)/2;

        return Point3D(x_m, y_m, z_m)

    def distanceWithPoint(self, p):
        '''distanceWithPoint: Calculate distance with Point3D
        Args:
            p (Point3D): The other point.

        Returns:
            (float): The distance between two points.
        '''

        return np.sqrt((p.x-self.x)**2 + (p.y-self.y)**2 + (p.z-self.z)**2)

    def scale(self, lower_bound=(0,0,0), scale_ratio=1, base=0):
        '''scale: Scale the sphere with respect to lower_bound with scale_ratio.
        Args:
            lower_bound ((float, float, float)): The lower_bound.
            scale_ratio (float): The scale ratio.
            base (float): New base
            ----------------------------------------
            new_x = (x-lower_bound[0])*scale_ratio + base
            new_y = (y-lower_bound[1])*scale_ratio + base
            new_z = (z-lower_bound[2])*scale_ratio + base
        '''
        self.x = (self.x-lower_bound[0])*scale_ratio + base
        self.y = (self.y-lower_bound[1])*scale_ratio + base
        self.z = (self.z-lower_bound[2])*scale_ratio + base

class Obj3D():
    """ Base 3D Object Class
    """
    def __init__(self, p):
        '''Initialization function.
        Args:
            p (Point3D): Center of object.
        '''
        self.center =  p

    def __str__(self):
        return 'Obj3D: position {}'.format(self.center)

    def scale(self, lower_bound=(0,0,0), scale_ratio=1, base=0):
        '''scale: Scale the sphere with respect to lower_bound with scale_ratio.
        Args:
            lower_bound ((float, float, float)): The lower_bound.
            scale_ratio (float): The scale ratio.
            base (float): New base
        '''
        self.center.scale(lower_bound, scale_ratio, base)

class Sphere(Obj3D):
    """ Sphere inherit from Obj3D
    """
    def __init__(self, center_point, radius):
        '''__init__:
        Args:
            center_point (Obj3D): Center point.
            radius (float): Radius of sphere.
        '''
        self.center_point = center_point
        self.radius = radius
        super().__init__(center_point)

    def __str__(self):
        return 'Sphere:\n' \
               +'\tcenter_point: '+self.center_point.__str__()+'\n' \
               +'\tradius:{}'.format(self.radius)

    def calBBox(self):
            min_mat = np.array(self.center_point.toList()) # 1*3
            max_mat = min_mat.copy()
            r_mat = np.array([self.radius]) # 1*1
            r_mat = np.tile(r_mat, (1,3))

            min_mat = min_mat-r_mat
            max_mat = max_mat+r_mat+1

            min_bound = np.floor(np.min(min_mat, 0)).astype(np.int)
            max_bound = np.ceil(np.max(max_mat, 0)).astype(np.int)
            return tuple(np.hstack((min_bound, max_bound)))

    def scale(self, lower_bound=(0,0,0), scale_ratio=1, base=0):
        '''scale: Scale the sphere with respect to lower_bound with scale_ratio.
        Args:
            lower_bound ((float, float, float)): The lower_bound.
            scale_ratio (float): The scale ratio.
            base (float): New base
            ----------------------------------------
        '''
        self.center_point.scale(lower_bound, scale_ratio, base)
        self.radius = self.radius*scale_ratio


class Cone(Obj3D):
    """ Cone inherit from Obj3D
    """
    def __init__(self, bottom_point, bottom_radius, up_point, up_radius):
        '''__init__:
        Args:
            bottom_point (Obj3D): Bottom point.
            bottom_radius (float): Bottom radius.
            up_point (Obj3D): Up point.
            up_radius (float): Up radius.
        '''
        if bottom_radius<up_radius:
            self.bottom_point =  up_point
            self.bottom_radius = up_radius
            self.up_point = bottom_point
            self.up_radius = bottom_radius
        else:
            self.bottom_point = bottom_point
            self.bottom_radius = bottom_radius
            self.up_point = up_point
            self.up_radius = up_radius

        self.height=bottom_point.distanceWithPoint(up_point)
        super().__init__(bottom_point.medianWithPoint(up_point))

    def __str__(self):
            return 'Cone:\n' \
                   +'\tbottom_point: '+self.bottom_point.__str__()+'\n' \
                   +'\tbottom_radius: {}'.format(self.bottom_radius)+'\n' \
                   +'\tup_point: '+self.up_point.__str__()+'\n' \
                   +'\tup_radius: {}'.format(self.up_radius)+'\n' \
                   +'\theight: {}'.format(self.height)

    def calBBox(self):
            min_mat = np.array([self.up_point.toList(), self.bottom_point.toList()])
            max_mat = min_mat.copy()
            r_mat = np.array([[self.up_radius], [self.bottom_radius]])
            r_mat = np.tile(r_mat, (1,3))

            min_mat = min_mat-r_mat
            max_mat = max_mat+r_mat+1

            min_bound = np.floor(np.min(min_mat, 0)).astype(np.int)
            max_bound = np.ceil(np.max(max_mat, 0)).astype(np.int)
            return tuple(np.hstack((min_bound, max_bound)))

    def revertMat(self):
        '''revertMat: Calculate revert matrix.
        Returns:
            bool: The return value.
        '''
        p1 = self.bottom_point
        p2 = self.up_point
        p3=p1.medianWithPoint(p2)
        v0=np.array([[p1.x, p2.x, p3.x],
                     [p1.y, p2.y, p3.y],
                     [p1.z, p2.z, p3.z],
                     [   1,    1,    1]])
        Dis=p1.distanceWithPoint(p2)
        v1=np.array([[0,     0,   0],
                     [0,     0,   0],
                     [0, Dis, Dis/2],
                     [1,     1,  1]])
        rev_mat = superimposition_matrix(v0,v1)
        return rev_mat

    def scale(self, lower_bound=(0,0,0), scale_ratio=1, base=0):
        '''scale: Scale the sphere with respect to lower_bound with scale_ratio.
        Args:
            lower_bound ((float, float, float)): The lower_bound.
            scale_ratio (float): The scale ratio.
            base (float): New base
            ----------------------------------------
        '''
        self.bottom_point.scale(lower_bound, scale_ratio, base)
        self.bottom_radius = self.bottom_radius*scale_ratio
        self.up_point.scale(lower_bound, scale_ratio, base)
        self.up_radius = self.up_radius*scale_ratio
        self.height = self.bottom_point.distanceWithPoint(self.up_point)

def calculateBound(sphere_list):
    '''calculateBound: Calculate the bound of spheres.
    Args:
        sphere_list (list of Sphere).
    Returns:
        min_bound ([float, float, float]): minimum of x,y,z.
        max_bound ([float, float, float]): maximum of x,y,z.
    '''
    min_mat = np.array([p.center_point.toList() for p in sphere_list])
    max_mat = min_mat.copy()
    r_mat = np.array([p.radius for p in sphere_list])
    r_mat = np.transpose(np.tile(r_mat, (3,1)))

    min_mat = min_mat-r_mat
    max_mat = max_mat+r_mat

    min_bound = np.min(min_mat, 0)
    max_bound = np.max(max_mat, 0)

    return min_bound, max_bound

def calScaleRatio(min_bound, max_bound, img_lengths):
    '''calScaleRatio: Calculate the scale ratio with respect to length.
    Args:
        min_bound ([float, float, float]): minimum of x,y,z.
        max_bound ([float, float, float]): maximum of x,y,z.
        img_lengths ((float, float, float)): The desired length.
    Returns:
        (float): The calculated ratio.
    '''
    spans = max_bound-min_bound
    scale_ratios = np.array(img_lengths)/np.array(spans)
    return np.min(scale_ratios)



import queue
import numpy as np
from pyneval.model.swc_node import SwcNode, SwcTree, Make_Virtual
from pyneval.model.euclidean_point import Line, EuclideanPoint
from anytree import PreOrderIter
from pyneval.io.swc_writer import swc_save
import copy


def itp_ok(node=None, son=None, pa=None,
           rad_mul=1.50, center_dis=None):
    '''
    judge if node can be canceled
    :param node: main node, may be deleted
    :param son: node's son
    :param pa: node's parent
    :param rad_mul:
    :param center_dis:
    :return:
    '''
    f_line = Line(e_node_1=son.get_center(), e_node_2=pa.get_center())
    e_node = node.get_center()

    foot = e_node.get_foot_point(f_line)
    if not foot.on_line(f_line):
        return False
    else:
        itp_dis = e_node.distance(foot)
        e_son = son._pos
        e_pa = pa._pos
        itp_rad = son.radius() + (pa.radius() - son.radius()) * e_son.distance(foot) / e_son.distance(e_pa)
        if center_dis is None:
            center_dis = node.radius()/2
        if node.radius() / itp_rad > rad_mul or itp_dis > center_dis:
            return False
    return True


def down_sample(swc_tree=None, rad_mul=1.50, center_dis=None, stage=0, k=1.0):
    stack = queue.LifoQueue()
    stack.put(swc_tree.root())
    down_pa = {}
    is_active = [True]*(max(swc_tree.id_set) + 5)

    for node in PreOrderIter(swc_tree.root()):
        if node.parent is None or node is None:
            continue
        down_pa[node] = node.parent

    while not stack.empty():
        node = stack.get()
        for son in node.children:
            stack.put(son)

        # 确保不是根节点
        if node.is_virtual() or down_pa[node].is_virtual():
            continue
        # 确保是二度节点
        if len(node.children) != 1:
            continue

        son, pa = node.children[0], down_pa[node]
        son_dis, pa_dis, grand_dis = son.parent_distance(), node.distance(down_pa[node]), son.distance(pa)

        # 确保针对采样率高的情况
        if stage == 0 and (son_dis > k*(son.radius() + node.radius()) or pa_dis > k*(pa.radius() + node.radius())):
            continue
        if stage == 1 and (son_dis > k*(son.radius() + node.radius()) and pa_dis > k*(pa.radius() + node.radius())):
            continue
        if itp_ok(node=node, son=son, pa=pa, rad_mul=rad_mul, center_dis=center_dis):
            is_active[node.get_id()] = False
            down_pa[son] = down_pa[node]

    return down_pa, is_active


def down_sample_swc_tree_command_line(swc_tree, config=None):
    rad_mul = config['rad_mul']
    center_dis = config['center_dis']
    stage = config['stage']
    return down_sample_swc_tree_stage(swc_tree=swc_tree, rad_mul=rad_mul, center_dis=center_dis, stage=stage)


def down_sample_swc_tree_stage(swc_tree, rad_mul=1.50, center_dis=None, stage=0, k=1.0):
    '''
    :param swc_tree: the tree need to delete node
    :param rad_mul: defult=1.5
    :param center_dis: defult=None
    :param stage: 0: for 2 degree node, delete if one side is too close, 1: for 2 degree node, delete if two sides are two close
    :return: swc_tree has changed in this function
    '''
    down_pa, is_activate = down_sample(swc_tree=swc_tree, rad_mul=rad_mul, center_dis=center_dis, stage=stage, k=k)
    new_swc_tree = SwcTree()
    node_list = [node for node in PreOrderIter(swc_tree.root())]
    id_node_map = {-1: new_swc_tree.root()}
    for node in node_list:
        if node.is_virtual():
            continue
        if is_activate[node.get_id()]:
            tmp_node = SwcNode()
            tmp_node._id = node.get_id()
            tmp_node._type = node._type
            tmp_node._pos = copy.copy(node._pos)
            tmp_node._radius = node._radius
            pa = id_node_map[down_pa[node].get_id()]
            tmp_node.parent = pa
            id_node_map[node.get_id()] = tmp_node
            new_swc_tree.id_set.add(tmp_node._id)
    return new_swc_tree


def down_sample_swc_tree(swc_tree, rad_mul=1.50, center_dis=None, k=1.0):
    res1 = down_sample_swc_tree_stage(swc_tree, rad_mul=rad_mul, center_dis=center_dis, stage=1, k=k)
    res2 = down_sample_swc_tree_stage(res1, rad_mul=rad_mul, center_dis=center_dis, stage=0, k=k)
    return res2


def up_sample_swc_tree_command_line(swc_tree, config=None):
    length_threshold = float(config['length_threshold'])
    return up_sample_swc_tree(swc_tree=swc_tree, length_threshold=length_threshold)


def up_sample_swc_tree(swc_tree, length_threshold):
    '''
    :param swc_tree: the tree need to upsample
    :param length_threshold: the distance between upsampled adjacent nodes.
                             floor(length_of_length/threshold - 1) new nodes will be interpolated in to an edge.
                             if length_of_length < threshold, no new nodes will be added.
    :return: upsampled swc tree
    '''
    up_sampled_swc_tree = swc_tree.get_copy()
    node_list = up_sampled_swc_tree.get_node_list()
    for node in node_list:
        # actually only div_num - 1 new nodes will be interpolated, id from 1 to div_num-1
        div_num = int(np.floor(node.parent_distance() / length_threshold))
        pa = node.parent
        for i in range(1, div_num):
            new_pos = EuclideanPoint(center=[node.get_x() + (pa.get_x() - node.get_x())/div_num*i,
                                             node.get_y() + (pa.get_y() - node.get_y())/div_num*i,
                                             node.get_z() + (pa.get_z() - node.get_z())/div_num*i])

            new_node = SwcNode(center=new_pos)
            new_node.set_r(node.radius() + (pa.radius() - node.radius())/div_num*i)
            new_node._type = 7

            up_sampled_swc_tree.unlink_child(node)
            if not up_sampled_swc_tree.add_child(pa, new_node):
                raise Exception("[Error: ] add child fail type of pa :{}, type of son".format(type(pa, new_node)))
            if not up_sampled_swc_tree.link_child(new_node, node):
                raise Exception("[Error: ] add child fail type of pa :{}, type of son".format(type(new_node, node)))
    up_sampled_swc_tree.get_node_list(update=True)
    return up_sampled_swc_tree



import tflib as lib

import numpy as np
import tensorflow as tf

def Batchnorm(name, axes, inputs, is_training=None, stats_iter=None, update_moving_stats=True, fused=True):
    if ((axes == [0,2,3]) or (axes == [0,2])) and fused==True:
        if axes==[0,2]:
            inputs = tf.expand_dims(inputs, 3)
        # Old (working but pretty slow) implementation:
        ##########

        # inputs = tf.transpose(inputs, [0,2,3,1])

        # mean, var = tf.nn.moments(inputs, [0,1,2], keep_dims=False)
        # offset = lib.param(name+'.offset', np.zeros(mean.get_shape()[-1], dtype='float32'))
        # scale = lib.param(name+'.scale', np.ones(var.get_shape()[-1], dtype='float32'))
        # result = tf.nn.batch_normalization(inputs, mean, var, offset, scale, 1e-4)

        # return tf.transpose(result, [0,3,1,2])

        # New (super fast but untested) implementation:
        offset = lib.param(name+'.offset', np.zeros(inputs.get_shape()[1], dtype='float32'))
        scale = lib.param(name+'.scale', np.ones(inputs.get_shape()[1], dtype='float32'))

        moving_mean = lib.param(name+'.moving_mean', np.zeros(inputs.get_shape()[1], dtype='float32'), trainable=False)
        moving_variance = lib.param(name+'.moving_variance', np.ones(inputs.get_shape()[1], dtype='float32'), trainable=False)

        def _fused_batch_norm_training():
            return tf.nn.fused_batch_norm(inputs, scale, offset, epsilon=1e-5, data_format='NCHW')
        def _fused_batch_norm_inference():
            # Version which blends in the current item's statistics
            batch_size = tf.cast(tf.shape(inputs)[0], 'float32')
            mean, var = tf.nn.moments(inputs, [2,3], keep_dims=True)
            mean = ((1./batch_size)*mean) + (((batch_size-1.)/batch_size)*moving_mean)[None,:,None,None]
            var = ((1./batch_size)*var) + (((batch_size-1.)/batch_size)*moving_variance)[None,:,None,None]
            return tf.nn.batch_normalization(inputs, mean, var, offset[None,:,None,None], scale[None,:,None,None], 1e-5), mean, var

            # Standard version
            # return tf.nn.fused_batch_norm(
            #     inputs,
            #     scale,
            #     offset,
            #     epsilon=1e-2, 
            #     mean=moving_mean,
            #     variance=moving_variance,
            #     is_training=False,
            #     data_format='NCHW'
            # )

        if is_training is None:
            outputs, batch_mean, batch_var = _fused_batch_norm_training()
        else:
            outputs, batch_mean, batch_var = tf.cond(is_training,
                                                       _fused_batch_norm_training,
                                                       _fused_batch_norm_inference)
            if update_moving_stats:
                no_updates = lambda: outputs
                def _force_updates():
                    """Internal function forces updates moving_vars if is_training."""
                    float_stats_iter = tf.cast(stats_iter, tf.float32)

                    update_moving_mean = tf.assign(moving_mean, ((float_stats_iter/(float_stats_iter+1))*moving_mean) + ((1/(float_stats_iter+1))*batch_mean))
                    update_moving_variance = tf.assign(moving_variance, ((float_stats_iter/(float_stats_iter+1))*moving_variance) + ((1/(float_stats_iter+1))*batch_var))

                    with tf.control_dependencies([update_moving_mean, update_moving_variance]):
                        return tf.identity(outputs)
                outputs = tf.cond(is_training, _force_updates, no_updates)

        if axes == [0,2]:
            return outputs[:,:,:,0] # collapse last dim
        else:
            return outputs
    else:
        # raise Exception('old BN')
        # TODO we can probably use nn.fused_batch_norm here too for speedup
        mean, var = tf.nn.moments(inputs, axes, keep_dims=True)
        shape = mean.get_shape().as_list()
        if 0 not in axes:
            print("WARNING ({}): didn't find 0 in axes, but not using separate BN params for each item in batch".format(name))
            shape[0] = 1
        offset = lib.param(name+'.offset', np.zeros(shape, dtype='float32'))
        scale = lib.param(name+'.scale', np.ones(shape, dtype='float32'))
        result = tf.nn.batch_normalization(inputs, mean, var, offset, scale, 1e-5)


        return result



import argparse
import os
import platform
import sys

from pyneval.errors.exceptions import InvalidMetricError
from pyneval.io.read_json import read_json
from pyneval.io.read_swc import read_swc_trees
from pyneval.io.swc_writer import swc_save
from pyneval.tools.overlap_detect import overlap_clean
from pyneval.tools.re_sample import down_sample_swc_tree_command_line, up_sample_swc_tree_command_line

method_list = ["up_sample", "down_sample", "overlap_clean", "US", "DS", "OC"]


def read_parameters():
    parser = argparse.ArgumentParser(description="pyneval 1.0")
    parser.add_argument("--swc_file", "-S", help="the route of the test file", required=False, nargs="*")
    parser.add_argument("--method", "-M", help="choose a adjust method", required=True)
    parser.add_argument(
        "--output", "-O", help="the route of the output file.\nif not specified, output to screen", required=True
    )
    parser.add_argument("--config", "-C", help="special config for different metric method", required=False)
    parser.add_argument("--debug", "-D", help="Print debug info or not", required=False)
    return parser.parse_args()


def run():
    # init path parameter
    abs_dir = os.path.abspath("")
    sys.path.append(abs_dir)
    sys.path.append(os.path.join(abs_dir, "src"))
    sys.path.append(os.path.join(abs_dir, "test"))
    sys.setrecursionlimit(1000000)

    # read parameter
    args = read_parameters()

    # file paths
    if args.swc_file is None:
        swc_files = []
    else:
        swc_files = [os.path.join(abs_dir, path) for path in args.swc_file]

    # method
    method = args.method
    if method not in method_list:
        raise InvalidMetricError(method)

    # out_path
    out_path = args.output
    # config
    config = args.config
    if config is None:
        if platform.system() == "Windows":
            if method in ["up_sample", "US"]:
                config = os.path.join(abs_dir, "config\\up_sample.json")
            if method in ["down_sample", "DS"]:
                config = os.path.join(abs_dir, "config\\length_metric.json")
            if method in ["overlap_clean", "OC"]:
                config = os.path.join(abs_dir, "config\\overlap_clean.json")
        elif platform.system() == "Linux":
            if method == ["up_sample", "US"]:
                config = os.path.join(abs_dir, "config/diadem_metric.json")
            if method in ["overall_length", "matched_length", "OL", "ML"]:
                config = os.path.join(abs_dir, "config/length_metric.json")
            if method in ["overlap_clean", "OC"]:
                config = os.path.join(abs_dir, "config/overlap_clean.json")
    config = read_json(json_file_path=config)
    tree_name_dict = {}
    swc_trees = []
    for file in swc_files:
        swc_trees += read_swc_trees(file, tree_name_dict=tree_name_dict)

    it = 1
    for swc_tree in swc_trees:
        if method in ["up_sample", "US"]:
            up_sampled_swc_tree = up_sample_swc_tree_command_line(swc_tree=swc_tree, config=config)
            swc_save(
                swc_tree=up_sampled_swc_tree, out_path=os.path.join(out_path, "up_sampled_" + tree_name_dict[swc_tree])
            )
        if method in ["down_sample", "DS"]:
            config["stage"] = 0
            down_sampled_swc_tree = down_sample_swc_tree_command_line(swc_tree, config=config)
            config["stage"] = 1
            down_sampled_swc_tree = down_sample_swc_tree_command_line(down_sampled_swc_tree, config=config)
            swc_save(
                swc_tree=down_sampled_swc_tree,
                out_path=os.path.join(out_path, "down_sampled_" + tree_name_dict[swc_tree]),
            )
        if method in ["overlap_clean", "OC"]:
            overlap_clean(swc_tree, out_path, tree_name_dict[swc_tree], config)
        it += 1


if __name__ == "__main__":
    sys.exit(run())


# python ./pyneval_tools.py
# --swc_file D:\gitProject\mine\PyNeval\test\data_example\gold\swc_18254_1
# --config D:\gitProject\mine\PyNeval\config\overlap_clean.json
# --method overlap_clean
# --output D:\\gitProject\\mine\\PyNeval\\output\\overlap_clean

# pyneval_tools --swc_file D:\gitProject\mine\PyNeval\test\data_example\gold\swc_18254_1 --config D:\gitProject\mine\PyNeval\config\overlap_clean.json --method overlap_clean --output D:\\gitProject\\mine\\PyNeval\\output\\overlap_clean

# pyneval_tools --swc_file D:\gitProject\mine\PyNeval\test\data_example\gold\2_18_gold.swc --config D:\gitProject\mine\PyNeval\config\down_sample.json --method down_sample --output D:\\gitProject\\mine\\PyNeval\\output\\down_sample

# pyneval_tools --swc_file D:\gitProject\mine\PyNeval\test\data_example\test\2_18_test.swc --config D:\gitProject\mine\PyNeval\config\up_sample.json --method up_sample --output D:\\gitProject\\mine\\PyNeval\\output\\up_sample



import json
import os
import sys

from pyneval.errors.exceptions import PyNevalError


def is_json_file(json_file):
    return json_file[-5:] in (".json", ".JSON")


def read_json(json_file_path, DEBUG=False):
    json_file_path = os.path.normpath(json_file_path)
    with open(json_file_path, "r") as f:
        data = json.load(f)
        if DEBUG:
            print (type(data))
        return data


def save_json(json_file_path, data, DEBUG=False):
    json_file_path = os.path.normpath(json_file_path)
    json_file_dir = os.path.dirname(json_file_path)
    while not os.path.exists(json_file_dir):
        print('[Info]: "{}" dose not exist. Create new path?[y/n]'.format(json_file_dir))
        choice = input()
        if choice.lower() == 'y':
            os.makedirs(json_file_dir)
        elif choice.lower() == 'n':
            print('[Info]: "{}" dose not exist. Input new directory?[y/n]'.format(json_file_dir))
            choice = input()
            if choice.lower() == 'y':
                print('[Info]: Input new directory')
                json_file_dir = input()
            elif choice.lower() == 'n':
                return False
            else:
                continue
    json_file_new_path = os.path.join(json_file_dir, os.path.basename(json_file_path))
    while not is_json_file(json_file_new_path):
        print('[Info]: "{}"  is not a json file. Input another path?[y/n]'.format(json_file_new_path))
        choice = input()
        if choice.lower() == 'y':
            print('[Info]: Input new path:')
            json_file_new_path = input()
        elif choice.lower() == 'n':
            return False

    while os.path.exists(json_file_new_path):
        print('[Info]: "{}" is already existed. Overwrite?[y/n]'.format(json_file_new_path))
        choice = input()
        if choice.lower() == 'y':
            break
        elif choice.lower() == 'n':
            return False

    with open(json_file_new_path, "w") as f:
        json.dump(data, f, indent=4)
        if DEBUG:
            print(type(data))
    return True




class PyNevalError(Exception):
    default_detail = "ERROR: PyNeval Error"

    def __init__(self, detail=None):
        if detail is None:
            detail = self.default_detail
        self.detail = "ERROR: {}".format(detail)

    def __str__(self):
        return self.detail


class InvalidSwcFileError(PyNevalError):
    """
    invalid swc file
    """
    default_detail = "ERROR: Invalid swc file"
    def __init__(self, file=None):
        if file is not None:
            self.detail = "ERROR: {} is a invalid swc file".format(file)
        else:
            self.detail = self.default_detail


class InvalidTiffFileError(PyNevalError):
    """
    invalid swc file
    """
    default_detail = "ERROR: Invalid tiff file"
    def __init__(self, file=None):
        if file is not None:
            self.detail = "ERROR: {} is a invalid tiff file".format(file)
        else:
            self.detail = self.default_detail


class InvalidMetricError(PyNevalError):
    """
    invalid metric
    """
    default_detail = "ERROR: The metric is not supported."
    def __init__(self, metric=None, valid_metrics=None):
        self.detail = self.default_detail
        if metric is not None:
            self.detail = "ERROR: The metric '{}' is not supported.".format(metric)
        if valid_metrics is not None:
            self.detail += "\nValid options for --metric:\n{}".format(valid_metrics)

class InvalidEuclideanPoint(PyNevalError):
    """
    Invalid Euclidean point
    """
    default_detail = "ERROR: invalid euclidean point"

    def __init__(self, param=None):
        self.detail = self.default_detail
        if param is not None:
            self.detail = "ERROR: {} is not a valid eculidean point".format(param)

class InvalidNode(PyNevalError):
    """
    Invalid node
    """
    default_detail = "ERROR: invalid node"

    def __init__(self, detail=""):
        self.detail = self.default_detail
        if detail:
            self.detail = "ERROR: {}".format(detail)


import unittest, time
from pyneval.io.read_json import read_json
from pyneval.model.swc_node import SwcNode, SwcTree
from pyneval.metric.length_metric import length_metric


class LengthMetricTest(unittest.TestCase):
    def test_len_1(self):
        goldtree = SwcTree()
        testTree = SwcTree()

        goldtree.load("D:\gitProject\mine\PyNeval\\test\data_example\gold\\2_18_gold.swc")
        testTree.load("D:\gitProject\mine\PyNeval\\test\data_example\\test\\2_18_test.swc")

        start = time.time()
        lm_res = length_metric(gold_swc_tree=goldtree,
                               test_swc_tree=testTree,
                               config=read_json("D:\gitProject\mine\PyNeval\config\length_metric.json"))

        lm_res2 = length_metric(gold_swc_tree=testTree,
                                test_swc_tree=goldtree,
                                config=read_json("D:\gitProject\mine\PyNeval\config\length_metric.json"))

        print("2_18 {}".format(time.time()-start))
        self.assertEqual(lm_res[0], 0.74957484)
        self.assertEqual(lm_res2[0], 0.91178429)
        self.assertEqual(lm_res[1], 0.77636927)
        self.assertEqual(lm_res2[1], 0.84352877)

    def test_len_2(self):
        goldtree = SwcTree()
        testTree = SwcTree()

        goldtree.load("D:\gitProject\mine\PyNeval\\test\data_example\gold\\30_18_10_gold.swc")
        testTree.load("D:\gitProject\mine\PyNeval\\test\data_example\\test\\30_18_10_test.swc")

        start = time.time()
        recall1, precision1, vertical_tree = length_metric(gold_swc_tree=goldtree,
                                                           test_swc_tree=testTree,
                                                           config=read_json("D:\gitProject\mine\PyNeval\config\length_metric.json"))

        recall2, precision2, vertical_tree = length_metric(gold_swc_tree=testTree,
                                                           test_swc_tree=goldtree,
                                                           config=read_json("D:\gitProject\mine\PyNeval\config\length_metric.json"))
        print("30_18_10 {}".format(time.time()-start))
        self.assertEqual(recall1, 0.95037505)
        self.assertEqual(recall2, 0.88408952)
        self.assertEqual(precision1, 0.88201204)
        self.assertEqual(precision2, 0.95769503)

    def test_len_3(self):
        goldtree = SwcTree()
        testTree = SwcTree()

        goldtree.load("D:\gitProject\mine\PyNeval\\test\data_example\gold\\34_23_10_gold.swc")
        testTree.load("D:\gitProject\mine\PyNeval\\test\data_example\\test\\34_23_10_test.swc")

        start = time.time()
        recall1, precision1, vertical_tree = length_metric(gold_swc_tree=goldtree,                                                    
                                                           test_swc_tree=testTree,                                                    
                                                           config=read_json("D:\gitProject\mine\PyNeval\config\length_metric.json"))
                                                                                                                                      
        recall2, precision2, vertical_tree = length_metric(gold_swc_tree=testTree,                                                    
                                                           test_swc_tree=goldtree,                                                    
                                                           config=read_json("D:\gitProject\mine\PyNeval\config\length_metric.json"))
        print("34_23_10 {}".format(time.time()-start))
        self.assertEqual(recall1, 0.66041513)
        self.assertEqual(recall2, 0.60889348)
        self.assertEqual(precision1, 0.58476999)
        self.assertEqual(precision2, 0.70757075)

    def test_len_4(self):
        goldtree = SwcTree()
        testTree = SwcTree()

        goldtree.load("D:\gitProject\mine\PyNeval\\test\data_example\gold\\conner.swc")
        testTree.load("D:\gitProject\mine\PyNeval\\test\data_example\\test\\conner.swc")

        recall1, precision1, vertical_tree = length_metric(gold_swc_tree=goldtree,                                                    
                                                           test_swc_tree=testTree,                                                    
                                                           config=read_json("D:\gitProject\mine\PyNeval\config\length_metric.json"))
                                                                                                                                      
        recall2, precision2, vertical_tree = length_metric(gold_swc_tree=testTree,                                                    
                                                           test_swc_tree=goldtree,                                                    
                                                           config=read_json("D:\gitProject\mine\PyNeval\config\length_metric.json"))

        self.assertEqual(recall1, 1.0)
        self.assertEqual(recall2, 0.0)
        self.assertEqual(precision1, 1.0)
        self.assertEqual(precision2, 0.0)

    def test_len_5(self):
        goldtree = SwcTree()
        testTree = SwcTree()

        goldtree.load("D:\gitProject\mine\PyNeval\\test\data_example\gold\multy_useage\push.swc")
        testTree.load("D:\gitProject\mine\PyNeval\\test\data_example\\test\multy_useage\push.swc")

        recall1, precision1, vertical_tree = length_metric(gold_swc_tree=goldtree,
                                                           test_swc_tree=testTree,
                                                           config=read_json(
                                                               "D:\gitProject\mine\PyNeval\config\length_metric.json"))

        recall2, precision2, vertical_tree = length_metric(gold_swc_tree=testTree,
                                                           test_swc_tree=goldtree,
                                                           config=read_json(
                                                               "D:\gitProject\mine\PyNeval\config\length_metric.json"))

        self.assertEqual(recall1, 0.76588382)
        self.assertEqual(recall2, 0.39212906)
        self.assertEqual(precision1, 0.40348057)
        self.assertEqual(precision2, 0.75791792)

    def test_len_6(self):
        test_tree = SwcTree()
        gold_tree = SwcTree()
        test_tree.load("../../../data/test_data/geo_metric_data/test/fake_data1.swc")
        gold_tree.load("../../../data/test_data/geo_metric_data/gold/fake_data1.swc")
        config = read_json("../../../config/length_metric.json")

        recall, precision = length_metric(gold_swc_tree=gold_tree, test_swc_tree=test_tree, config=config)
        print(recall)
        print(precision)


if __name__ == '__main__':
    unittest.main()



import queue
import time
import math
import numpy as np

from anytree import PreOrderIter
from pyneval.model import binary_node, swc_node
from pyneval.errors.exceptions import InvalidNode
from pyneval.metric.utils import diadam_match_utils, point_match_utils, bin_utils
from pyneval.metric.utils.metric_manager import get_metric_manager

metric_manager = get_metric_manager()

class DiademMetric(object):
    """
    Diadem metric
    """

    WEIGHT_DEGREE = 1
    WEIGHT_SQRT_DEGREE = 2
    WEIGHT_DEGREE_HARMONIC_MEAN = 3
    WEIGHT_PATH_LENGTH = 4
    WEIGHT_UNIFORM = 5
    _2D = "2d"
    _3D = "3d"

    def __init__(self, config, debug=False):
        self.DEBUG = debug

        # thresholds
        self.g_terminal_threshold = 0

        self.g_weight_sum = 0
        self.g_score_sum = 0
        self.g_quantity_score_sum = 0
        self.g_quantity_score = 0
        self.g_direct_match_score = 0
        self.g_final_score = -1

        self.g_miss = set()
        self.g_spur_set = set()
        self.g_matches = {}
        self.g_weight_dict = {}
        self.g_excess_nodes = {}
        self.g_distance_match = []
        self.g_continuation = []

        # switch
        self.g_remove_spur = False
        self.g_align_tree_by_root = False
        self.g_count_excess_nodes = True
        self.g_weight_node = self.WEIGHT_DEGREE
        self.g_list_miss = False
        self.g_list_distant_matches = False
        self.g_list_continuations = False
        self.g_find_proper_root = False

        if "remove_spur" in config:
            self.g_remove_spur = config["remove_spur"]
        if "align_tree_by_root" in config.keys():
            self.g_align_tree_by_root = config["align_tree_by_root"]
        if "count_excess_nodes" in config.keys():
            self.g_count_excess_nodes = config["count_excess_nodes"]
        if "weight_node" in config.keys():
            self.g_weight_node = config["weight_node"]
        if "list_miss" in config.keys():
            self.g_list_miss = config["list_miss"]
        if "list_distant_matches" in config.keys():
            self.g_list_distant_matches = config["list_distant_matches"]
        if "list_continuations" in config.keys():
            self.g_list_continuations = config["list_continuations"]
        if "find_proper_root" in config.keys():
            self.g_find_proper_root = config["find_proper_root"]

    def remove_spurs(self, bin_root, threshold):
        spur_set = set()
        stack = queue.LifoQueue()

        if bin_root.has_children:
            stack.put(bin_root)

        while not stack.empty():
            node = stack.get()
            if node.has_children():
                stack.put(node.left_son)
                stack.put(node.right_son)
            else:
                distance = node.data.path_length
                if distance < threshold and not (node in self.g_matches):
                    both_children_spur = False
                    spur_set.add(node)
                    if node.get_side() == binary_node.RIGHT:
                        l_node = node.parent.left_son
                        if (
                            not l_node.has_children
                            and l_node.data.path_length < threshold
                            and not l_node in self.g_matches
                        ):
                            both_children_spur = True
                            spur_set.add(l_node)
                            ll_node = stack.get()
                            if ll_node != l_node:
                                raise InvalidNode("stack top is not current node's twin")

                    if not both_children_spur:
                        spur_set.add(node.parent)
        return spur_set

    def generate_node_weights(self, bin_root, bin_subroot, spur_set):
        init_stack = queue.LifoQueue()
        main_stack = queue.LifoQueue()
        degree_dict = {}
        for subroot in bin_subroot:
            init_stack.put(subroot)
            main_stack.put(subroot)

            while not init_stack.empty():
                node = init_stack.get()
                if node.has_children():
                    init_stack.put(node.left_son)
                    init_stack.put(node.right_son)
                    main_stack.put(node.left_son)
                    main_stack.put(node.right_son)

            while not main_stack.empty():
                node = main_stack.get()

                if self.g_weight_node == self.WEIGHT_UNIFORM:
                    self.g_weight_dict[node] = 1
                else:
                    if self.DEBUG:
                        print ("Determining weight for {}".format(node.data.id))

                    if node.is_leaf():
                        self.g_weight_dict[node] = 1
                        degree_dict[node] = 1
                    else:
                        degree = 0
                        if node.left_son.is_leaf() and node.right_son.is_leaf():
                            if node.left_son in spur_set or node.right_son in spur_set:
                                degree = 1
                            else:
                                degree = 2
                        elif node in spur_set:
                            if node.left_son.is_leaf():
                                degree += degree_dict[node.right_son]
                            else:
                                degree += degree_dict[node.left_son]
                        else:
                            degree += degree_dict[node.left_son]
                            degree += degree_dict[node.right_son]
                        degree_dict[node] = degree
                        if self.g_weight_node == self.WEIGHT_DEGREE:
                            self.g_weight_dict[node] = degree
                        elif self.g_weight_node == self.WEIGHT_SQRT_DEGREE:
                            self.g_weight_dict[node] = math.sqrt(degree)
                        elif self.g_weight_node == self.WEIGHT_DEGREE_HARMONIC_MEAN:
                            self.g_weight_dict[node] = 2.0 / (
                                1.0 / degree_dict[node.getLeft()] + 1.0 / degree_dict.get[node.getRight()]
                            )
                        elif self.g_weight_node == self.WEIGHT_PATH_LENGTH:
                            self.g_weight_dict[node] = node.data.path_length
                        if self.DEBUG:
                            print ("nodeId = {}, degree = {}".format(node.data._id, degree))
                            if node.parent is not None:
                                print ("pa = {}".format(node.parent.data._id))
                            print ("-----------")

    def is_diadem_match(self, gold_node, nearby_node, bin_gold_list, bin_test_list):
        length_diff = diadam_match_utils.get_match_path_length_difference(
            gold_node, nearby_node, bin_gold_list, bin_test_list
        )
        if self.DEBUG:
            print (
                "gold_node = {}, test_node = {}, length_diff = {}".format(
                    gold_node.data.get_id(), nearby_node.data.get_id(), length_diff
                )
            )
        return length_diff < 1

    def get_best_match(self, match_list, gold_node, bin_test_list):
        if self.DEBUG:
            print ("Determining best match")
        confirm_list = []

        gold_data = gold_node.data
        des_branch_left = {}
        trajactory_map = {}

        if gold_node.has_children():
            nearby_list = []
            path_length_map = {}
            path_length_map[gold_node] = swc_node.SwcNode()

            target_dis_list = []
            for test_node in match_list:
                target_distances = swc_node.SwcNode()
                target_distances.xy_path_length = gold_node.data.distance(test_node.data, self._2D)
                target_distances.z_path_length = gold_node.data.distance(test_node.data, self._2D)
                target_dis_list.append(target_distances)

            current_list = []
            current_list.append(gold_node.left_son)
            trajactory_map[gold_node.left_son] = gold_node.data.left_trajectory
            des_branch_left[gold_node.left_son] = True

            current_list.append(gold_node.right_son)
            trajactory_map[gold_node.right_son] = gold_node.data.right_trajectory
            des_branch_left[gold_node.right_son] = False

            while len(current_list) > 0 and len(confirm_list) == 0:
                next_list = []
                for des_node in current_list:
                    target_tra = trajactory_map[des_node]
                    if target_tra.get_x() == -1.0 or target_tra.get_z() == -1.0:
                        tmp_tra = diadam_match_utils.get_trajectory_for_path(gold_node, des_node)
                        if target_tra.get_x() == -1.0:
                            target_tra.set_x(tmp_tra.get_x())
                            target_tra.set_y(tmp_tra.get_y())
                        if target_tra.get_z() == -1.0:
                            target_tra.set_z(tmp_tra.get_z())
                    des_tra = des_node.data.parent_trajectory

                    if des_node.has_children():
                        next_list.append(des_node.left_son)
                        next_list.append(des_node.right_son)
                        trajactory_map[des_node.left_son] = target_tra
                        trajactory_map[des_node.right_son] = target_tra

                    gold_path_length = swc_node.SwcNode()
                    gold_path_length.add_length(des_node.data)
                    parent_path_length = path_length_map[des_node.parent]
                    gold_path_length.add_length(parent_path_length)
                    path_length_map[des_node] = gold_path_length

                    nearby_list = diadam_match_utils.get_nearby_node_list(
                        des_node, bin_test_list, check_previous_use=True, g_matches=self.g_matches
                    )

                    for test_node in nearby_list:
                        test_des_node = test_node
                        test_path_length = swc_node.SwcNode()

                        des_dis = swc_node.SwcNode()
                        des_dis.xy_path_length = des_node.data.distance(test_node.data, self._2D)
                        des_dis.z_path_length = math.fabs(des_node.data.get_z() - test_node.data.get_z())

                        while test_node.parent is not None:
                            test_path_length.add_length(test_node.data)
                            test_node = test_node.parent
                            tmp_list = []
                            for match in match_list:
                                if test_node == match:
                                    test_XY_path_length = (
                                        test_path_length.xy_path_length
                                        + diadam_match_utils.get_end_node_XY_dis_diff(
                                            gold_data, target_tra, test_node.data
                                        )
                                        + diadam_match_utils.get_end_node_XY_dis_diff(
                                            des_node.data, des_tra, test_des_node.data
                                        )
                                    )
                                    test_Z_path_length = (
                                        test_path_length.z_path_length
                                        + diadam_match_utils.get_end_node_Z_dis_diff(
                                            gold_data, target_tra, test_node.data
                                        )
                                        + diadam_match_utils.get_end_node_Z_dis_diff(
                                            des_node.data, des_tra, test_des_node.data
                                        )
                                    )

                                    if (
                                        diadam_match_utils.path_length_matches(
                                            gold_path_length, test_XY_path_length, test_Z_path_length
                                        )
                                        < 1
                                    ):
                                        confirm_list.append(match)
                                        tmp_list.append(match)

                            for node in tmp_list:
                                match_list.remove(node)
                            while len(tmp_list):
                                tmp_list.pop()

                current_list = nearby_list
        if len(confirm_list) == 1:
            return confirm_list[0]

        if len(confirm_list) == 0:
            confirm_list = match_list

        closest_dis = -1
        closest_match = None
        for test_node in confirm_list:
            dis = test_node.data.distance(gold_node.data)
            if closest_dis == -1 or dis < closest_dis:
                closest_dis = dis
                closest_match = test_node
        return closest_match

    def get_closest_match(self, gold_node, bin_gold_list, bin_test_list):
        best_match = None

        # Step1: find node within threshold
        nearby_list = diadam_match_utils.get_nearby_node_list(
            gold_node=gold_node, bin_test_list=bin_test_list, check_previous_use=True, g_matches=self.g_matches
        )
        if self.DEBUG:
            print ("nearby list of {}:".format(gold_node.data.get_id()))
            for node in nearby_list:
                print (node.data.get_id())
            print ("----END----")

        # Step2: find match node whose ancestor is also matched
        match_list = []
        for nearby_node in nearby_list:
            if self.is_diadem_match(gold_node, nearby_node, bin_gold_list, bin_test_list):
                if self.DEBUG:
                    print ("{} {}".format(gold_node.data.get_id(), nearby_node.data.get_id()))
                match_list.append(nearby_node)

        if len(match_list) == 1:
            best_match = match_list[0]
        elif len(match_list) >= 1:
            best_match = self.get_best_match(match_list, gold_node, bin_test_list)

        if self.DEBUG:
            if best_match is not None:
                print ("best match {} {}".format(gold_node.data.get_id(), best_match.data.get_id()))
            else:
                print ("best match {} None".format(gold_node.data.get_id()))
        return best_match

    def get_des_in_for_continuation(
        self, first_node, ancestor_node, ancestor_match, ancestor_trajectory, path_length_map, bin_test_list
    ):
        spe_ancestor_trajectory = ancestor_trajectory
        test_matches = []

        stack = queue.LifoQueue()
        stack.put(first_node)
        while not stack.empty():
            gold_node = stack.get()
            des_tra = gold_node.data.parent_trajectory
            # print("gold_node id = {}".format(gold_node.data.get_id()))
            if gold_node in self.g_matches:
                test_matches.append(self.g_matches[gold_node])
            else:
                test_matches = diadam_match_utils.get_nearby_node_list(
                    gold_node, bin_test_list, check_previous_use=False
                )

            prev_path_length = path_length_map[gold_node.parent]
            gold_path_length = swc_node.SwcNode()
            gold_path_length.path_length = gold_node.data.path_length + prev_path_length.path_length
            gold_path_length.xy_path_length = gold_node.data.xy_path_length + prev_path_length.xy_path_length
            gold_path_length.z_path_length = gold_node.data.z_path_length + prev_path_length.z_path_length
            path_length_map[gold_node] = gold_path_length

            for child_match in test_matches:
                test_path_length = swc_node.SwcNode()
                test_path_length.add_data(child_match.data)
                tmp_node = child_match.parent
                # print("child match id = {}".format(
                #     child_match.data.get_id()))
                done = tmp_node == ancestor_match
                while not done:
                    if tmp_node in path_length_map.keys():
                        prev_path_length = path_length_map[tmp_node]
                        test_path_length.add_length(prev_path_length)
                        done = True
                    else:
                        test_path_length.add_length(tmp_node.data)
                        tmp_node = tmp_node.parent
                        if tmp_node is None:
                            done = True

                    if tmp_node == ancestor_match:
                        done = True

                if tmp_node is None:
                    if self.DEBUG:
                        print ("[info]: descendant not match")
                else:
                    path_length_map[child_match] = test_path_length
                    if ancestor_trajectory.get_x() == -1.0 or ancestor_trajectory.get_z() == -1.0:
                        spe_ancestor_trajectory = diadam_match_utils.get_trajectory_for_path(ancestor_node, gold_node)
                        if ancestor_trajectory.get_x != -1.0:
                            spe_ancestor_trajectory.set_x(ancestor_trajectory.get_x())
                            spe_ancestor_trajectory.set_y(ancestor_trajectory.get_y())
                        if ancestor_trajectory.get_z != -1.0:
                            spe_ancestor_trajectory.set_z(ancestor_trajectory.get_z())

                    test_xy_path_length = (
                        test_path_length.xy_path_length
                        + diadam_match_utils.get_end_node_XY_dis_diff(
                            ancestor_node.data, spe_ancestor_trajectory, ancestor_match.data
                        )
                        + diadam_match_utils.get_end_node_XY_dis_diff(gold_node.data, des_tra, child_match.data)
                    )
                    test_z_path_length = (
                        test_path_length.z_path_length
                        + diadam_match_utils.get_end_node_Z_dis_diff(
                            ancestor_node.data, spe_ancestor_trajectory, ancestor_match.data
                        )
                        + diadam_match_utils.get_end_node_Z_dis_diff(gold_node.data, des_tra, child_match.data)
                    )

                    if (
                        diadam_match_utils.path_length_matches(
                            gold_path_length, test_xy_path_length, test_z_path_length
                        )
                        < 1
                    ):
                        return child_match

            if gold_node.has_children() and gold_node not in self.g_matches:
                stack.put(gold_node.left_son)
                stack.put(gold_node.right_son)

    def is_sub_continuation(
        self, gold_node, ancestor_node, ancestor_match, is_left, gold_path_length, bin_test_list, add_to_list
    ):
        if ancestor_match is None:
            return False

        path_length_map = {}
        path_length_map[gold_node] = gold_path_length

        if is_left:
            ancestor_tra = ancestor_node.data.left_trajectory
        else:
            ancestor_tra = ancestor_node.data.right_trajectory

        left_child_match = self.get_des_in_for_continuation(
            gold_node.left_son, ancestor_node, ancestor_match, ancestor_tra, path_length_map, bin_test_list
        )
        right_child_match = self.get_des_in_for_continuation(
            gold_node.right_son, ancestor_node, ancestor_match, ancestor_tra, path_length_map, bin_test_list
        )

        if left_child_match is not None and right_child_match is not None:
            common_ancestor = diadam_match_utils.LCA(left_child_match, right_child_match, ancestor_match)
            if common_ancestor is not None and diadam_match_utils.is_within_dis_match_threshold(
                common_ancestor, gold_node
            ):
                self.g_matches[gold_node] = common_ancestor
                self.g_matches[common_ancestor] = gold_node
                if add_to_list:
                    self.g_distance_match.append(gold_node)
            elif add_to_list:
                self.g_continuation.append(gold_node)
            return True
        elif left_child_match is not None or right_child_match is not None:
            if add_to_list:
                self.g_continuation.append(gold_node)
            return True
        return False

    def is_continuation(self, gold_node, bin_test_list, add_to_list=True, DEBUG=False):
        if not gold_node.has_children():
            return False

        gold_path_length = swc_node.SwcNode()

        # find the first matched ancestor of the gold_node
        ancestor_gold = gold_node.parent
        gold_path_length.path_length = gold_node.data.path_length
        gold_path_length.xy_path_length = gold_node.data.xy_path_length
        gold_path_length.z_path_length = gold_node.data.z_path_length
        ancestor_match = False
        is_left = False

        if DEBUG:
            print (
                "gold_path_length: path_length {}, xy_path_length {}, z_path_length {}".format(
                    gold_path_length.path_length, gold_path_length.xy_path_length, gold_path_length.z_path_length
                )
            )
        while ancestor_gold is not None and not ancestor_match:
            if ancestor_gold in self.g_matches.keys():
                ancestor_match = self.g_matches[ancestor_gold]
                if self.is_sub_continuation(
                    gold_node, ancestor_gold, ancestor_match, is_left, gold_path_length, bin_test_list, add_to_list
                ):
                    return True
                ancestor_match = True
            ancestor_node_matches = diadam_match_utils.get_nearby_node_list(
                ancestor_gold, bin_test_list, check_previous_use=True, g_matches=self.g_matches
            )
            for _ in ancestor_node_matches:
                if self.is_sub_continuation(
                    gold_node, ancestor_gold, ancestor_match, is_left, gold_path_length, bin_test_list, add_to_list
                ):
                    return True

            gold_path_length.path_length += ancestor_gold.data.path_length
            gold_path_length.xy_path_length += ancestor_gold.data.xy_path_length
            gold_path_length.z_path_length += ancestor_gold.data.z_path_length
            is_left = ancestor_gold.is_left()
            ancestor_gold = ancestor_gold.parent
        # failed
        return False

    def get_excess_weight(self, excess, node, excess_map):
        if self.g_weight_node == self.WEIGHT_SQRT_DEGREE:
            return math.sqrt(excess)
        if self.g_weight_node == self.WEIGHT_DEGREE_HARMONIC_MEAN:
            return 2.0 / (1.0 / excess_map[node.left_son] + 1.0 / excess_map[node.right_son])
        if self.g_weight_node == self.WEIGHT_UNIFORM:
            return excess > 0
        if self.g_weight_node == self.WEIGHT_DEGREE:
            return excess

    def remove_nearest_node(self, node, nearby_nodes, bin_gold_list):
        closest_dis = -1
        closest_match = None

        for nearby_node in nearby_nodes:
            dis = node.data.distance(nearby_node.data)
            if closest_dis == -1 or dis < closest_dis:
                closest_dis = dis
                closest_match = nearby_node

        bin_gold_list.remove(closest_match)

    def weight_excess(self, bin_gold_subroot, bin_test_subroot):
        global g_excess_nodes

        if self.DEBUG:
            for node in self.g_matches.keys():
                print ("matches = {}".format(node.data.get_id()))

        sum_weight = 0.0
        gold_bin_list = []
        setup_stack = queue.LifoQueue()
        use_stack = queue.LifoQueue()

        for son in bin_gold_subroot:
            gold_bin_list.extend(son.get_node_list())

        for son in bin_test_subroot:
            setup_stack.put(son)
            use_stack.put(son)
            while not setup_stack.empty():
                node = setup_stack.get()
                if node.has_children():
                    setup_stack.put(node.left_son)
                    setup_stack.put(node.right_son)
                    use_stack.put(node.left_son)
                    use_stack.put(node.right_son)

            direct_term_excess = {}

            while not use_stack.empty():
                node = use_stack.get()
                data = node.data
                # if data.get_id()
                if self.DEBUG:
                    print ("{} sum weight {}".format(data.get_id(), sum_weight))

                if node.has_children():
                    num_excess_node = direct_term_excess[node.left_son]
                    num_excess_node += direct_term_excess[node.right_son]
                    excess_weight = self.get_excess_weight(num_excess_node, node, direct_term_excess)

                    if (
                        not node in self.g_matches.keys()
                        and len(
                            diadam_match_utils.get_nearby_node_list(
                                node, gold_bin_list, check_previous_use=True, g_matches=self.g_matches
                            )
                        )
                        == 0
                        and not self.is_continuation(node, gold_bin_list, add_to_list=False)
                    ):
                        self.g_excess_nodes[node] = excess_weight
                        if not node in self.g_spur_set:
                            sum_weight += excess_weight
                    else:
                        num_excess_node = 0
                        excess_weight = 0
                else:
                    num_excess_node = 0
                    excess_weight = 0

                    if (
                        node not in self.g_spur_set
                        and node not in self.g_matches.keys()
                        and node.parent not in self.g_matches.keys()
                    ):
                        nearby_nodes = diadam_match_utils.get_nearby_node_list(
                            node, gold_bin_list, check_previous_use=True, g_matches=self.g_matches
                        )
                        if len(nearby_nodes) > 0:
                            self.remove_nearest_node(node, nearby_nodes, gold_bin_list)
                        else:
                            num_excess_node = 1
                            excess_weight = 1
                            self.g_excess_nodes[node] = 1
                            sum_weight += excess_weight

                direct_term_excess[node] = num_excess_node
        return sum_weight

    def score_trees(self, bin_gold_root, bin_test_root, bin_gold_subroots, bin_test_subroots):
        # global g_weight_dict
        # global g_weight_sum
        # global g_score_sum
        # global g_quantity_score_sum
        # global g_final_score
        # global g_direct_match_score
        # global g_quantity_score

        stack = queue.LifoQueue()
        number_of_nodes = 0
        weight = 0

        bin_test_list = []
        bin_gold_list = []

        # Root is certainly matched, note that here is a address use
        bin_gold_root.data.parent_trajectory = bin_gold_root.data.get_center()
        bin_gold_root.data.left_trajectory = bin_gold_root.data.get_center()
        bin_gold_root.data.right_trajectory = bin_gold_root.data.get_center()
        if bin_test_root is not None:
            bin_test_root.data.parent_trajectory = bin_test_root.data.get_center()
            bin_test_root.data.left_trajectory = bin_test_root.data.get_center()
            bin_test_root.data.right_trajectory = bin_test_root.data.get_center()

            self.g_matches[bin_gold_root] = bin_test_root
            self.g_matches[bin_test_root] = bin_gold_root

            for son in bin_test_subroots:
                son.parent = bin_test_root
                bin_test_list.extend(son.get_node_list())

        for son in bin_gold_subroots:
            son.parent = bin_gold_root
            bin_gold_list.extend(son.get_node_list())

        # Increment quantity (non-continuations)
        self.g_quantity_score_sum += 1

        number_of_nodes -= len(self.g_spur_set)
        number_of_nodes += len(bin_gold_list)

        for gold_subroot in bin_gold_subroots:
            stack.put(gold_subroot)

            while not stack.empty():
                gold_node = stack.get()
                gold_data = gold_node.data
                if self.DEBUG:
                    print(gold_node.data.get_id())
                    if gold_node.data.get_id() in [4, 5, 6, 7]:
                        print("line:616")

                if gold_node.has_children():
                    stack.put(gold_node.left_son)
                    stack.put(gold_node.right_son)

                if gold_node in self.g_spur_set:
                    pass
                else:
                    weight = self.g_weight_dict[gold_node]
                    self.g_weight_sum += weight
                    match = self.get_closest_match(gold_node=gold_node, bin_gold_list=bin_gold_list,
                                              bin_test_list=bin_test_list)

                    # another match method (temporarily undo)
                    # if match is None and gold_node.is_leaf() and g_terminal_threshold > 0:
                    #     match = get_extended_termination_match(gold_node = gold_node,
                    #                                            bin_gold_list = bin_gold_list,
                    #                                            bin_test_list = bin_test_list)

                    if match is not None:
                        self.g_matches[gold_node] = match
                        self.g_matches[match] = gold_node
                        self.g_score_sum += weight

                        # Increment quantity (non-continuations)
                        self.g_quantity_score_sum += 1
                    else:
                        self.g_miss.add(gold_node)
            if self.DEBUG:
                gold_subroot.print_tree()
                print("--END--\n")
                if gold_subroot in self.g_matches.keys() is not None:
                    self.g_matches[gold_subroot].print_tree()
        t_remove = []

        for miss_node in self.g_miss:
            if miss_node.left_son is not None or miss_node.right_son is not None:
                if self.is_continuation(miss_node, bin_test_list):
                    weight = self.g_weight_dict[miss_node]
                    self.g_score_sum += weight
                    if self.DEBUG:
                        print("[Info: ] find continuation in miss nodes {}".format(miss_node.to_str()))
                    t_remove.append(miss_node)

        for node in t_remove:
            self.g_miss.remove(node)

        if self.g_weight_sum > 0:
            try:
                self.g_direct_match_score = self.g_quantity_score_sum / number_of_nodes
            except ZeroDivisionError:
                self.g_direct_match_score = 0
            try:
                self.g_quantity_score = self.g_score_sum / self.g_weight_sum
            except ZeroDivisionError:
                self.g_quantity_score = 0

            if self.g_remove_spur > 0:
                self.remove_spurs(bin_test_root)

            if self.g_count_excess_nodes:
                self.g_weight_sum += self.weight_excess(bin_gold_subroots, bin_test_subroots)

            try:
                self.g_final_score = self.g_score_sum / self.g_weight_sum
            except ZeroDivisionError:
                self.g_final_score = 0

    def color_tree_only(self):
        if self.g_list_miss:
            if len(self.g_miss) > 0:
                for node in self.g_miss:
                    # 9 means this node is missed
                    node.data._type = 9
            if len(self.g_excess_nodes) > 0:
                for node in self.g_excess_nodes.keys():
                    # 10 means this node is excessive
                    node.data._type = 10

        if self.g_list_continuations:
            if len(self.g_continuation) > 0:
                for node in self.g_continuation:
                    # 11 means this node is a continuation
                    node.data._type = 11

        if self.g_list_distant_matches:
            if len(self.g_distance_match) > 0:
                for node in self.g_distance_match:
                    # 12 means this node is a distant match
                    node.data._type = 12

    def print_result(self):
        start = time.time()
        print("g_weight_sum = {}".format(self.g_weight_sum))
        print("g_score_sum = {}".format(self.g_score_sum))
        print("g_final_score = {}".format(self.g_final_score))
        print("time cost = {}".format(time.time() - start))
        if self.g_list_miss:
            if len(self.g_miss) > 0:
                print("---Nodes that are missed (position and weight)---")
                for node in self.g_miss:
                    print(
                        "node_ID = {} poi = {} weight = {}".format(node.data.get_id(), node.data._pos,
                                                                   self.g_weight_dict[node])
                    )
                    # 9 means this node is missed
                    node.data._type = 9
                print("--END--")
            else:
                print("---Nodes that are missed:None---")

            print("")

            if len(g_excess_nodes) > 0:
                print("---extra Nodes in test reconstruction (position and weight)---")
                for node in g_excess_nodes.keys():
                    print(
                        "node_ID = {} poi = {} weight = {}".format(node.data.get_id(), node.data._pos,
                                                                   g_excess_nodes[node])
                    )
                    # 10 means this node is excessive
                    node.data._type = 10
            else:
                print("---extra Nodes in test reconstruction: None---")

        if self.g_list_continuations:
            print("")
            if len(self.g_continuation) > 0:
                print("---continuation Nodes (position and weight)---")
                for node in self.g_continuation:
                    print(
                        "node_ID = {} poi = {} weight = {}".format(node.data.get_id(), node.data._pos,
                                                                   self.g_weight_dict[node])
                    )
                    # 11 means this node is a continuation
                    node.data._type = 11
            else:
                print("---continuation Nodes None---")

        if self.g_list_distant_matches:
            if len(self.g_distance_match) > 0:
                print("Distant Matches")
                for node in self.g_distance_match:
                    print(
                        "node_ID = {} poi = {} weight = {}".format(node.data.get_id(), node.data._pos,
                                                                   self.g_weight_dict[node])
                    )
                    # 12 means this node is a distant match
                    node.data._type = 12
            else:
                print("Distant Matches: none")

    def adjust_root(self, swc_gold_tree, swc_test_tree, test_kdtree, test_pos_node_dict, t_matches):
        """
        find a common root for both gold and test trees
        Args:
            swc_gold_tree(Swc_Tree):
            swc_test_tree(Swc_Tree):
            test_kdtree(kd tree established based on test tree)
            test_pos_node_dict(Dict): a dict of test tree
                keys: tuple(x, y, z)
                value: Swc_Node in position of tuple
            t_matches(Dict):
                A match Dict of nodes in gold test
                key: Swc Node in gold tree
                value: Swc Node in test tree
        """
        swc_gold_list = swc_gold_tree.get_node_list()
        swc_test_list = swc_test_tree.get_node_list()

        gold_vis_list = np.zeros(shape=(len(swc_gold_list) + 10,))
        test_vis_list = np.zeros(shape=(len(swc_test_list) + 10,))

        for node in swc_gold_list:
            if node.is_virtual():
                continue
            nearby_nodes = swc_node.get_nearby_swc_node_list(
                gold_node=node, threshold=node.radius() / 2, test_kdtree=test_kdtree,
                test_pos_node_dict=test_pos_node_dict
            )
            for t_node in nearby_nodes:
                if not gold_vis_list[node.get_id()] and not test_vis_list[t_node.get_id()]:
                    t_matches[node] = t_node
                    swc_gold_tree.change_root(node.get_id())
                    swc_test_tree.change_root(t_node.get_id())

                    for sub_node in PreOrderIter(node):
                        gold_vis_list[sub_node.get_id()] = 1
                    for sub_node in PreOrderIter(t_node):
                        test_vis_list[sub_node.get_id()] = 1
                    break


@metric_manager.register(
    name="diadem",
    config="diadem_metric.json",
    desc="DIADEM metric (https://doi.org/10.1007/s12021-011-9117-y)",
    alias=["DM"],
    public=True,
)
def diadem_metric(gold_swc_tree, test_swc_tree, config, debug=False):
    """Main function of diadem metric
    Args:
        gold_swc_tree(SwcTree):
        test_swc_tree(SwcTree):
        config(Dict):
            The keys of 'config' is the name of configs, and the items are config values

    Example:
        test_tree = swc_node.SwcTree()
        gold_tree = swc_node.SwcTree()
        gold_tree.load("..\\..\\data\\test_data\\topo_metric_data\\sgold_fake_data2.swc")
        test_tree.load("..\\..\\data\\test_data\\topo_metric_data\\stest_fake_data2.swc")
        score, recall, precision = ssd_metric(gold_swc_tree=gold_tree,
                                              test_swc_tree=test_tree,
                                              config=config)

    Return:
        tuple: contain three values to demonstrate metric result
            g_weight_sum(float): weight of all nodes in the gold tree
            g_score_sum(float): weight of all the nodes that are matched.
            g_final_score(float): ratio of above two value

    Raise:
        None
    """

    gold_swc_tree.set_node_type_by_topo(root_id=1)
    test_swc_tree.set_node_type_by_topo(root_id=5)

    diadem = DiademMetric(config)
    diadam_match_utils.diadem_utils_init(config)
    scale = config["scale"]
    gold_swc_tree.rescale(scale)
    test_swc_tree.rescale(scale)
    test_kdtree, test_pos_node_dict = point_match_utils.create_kdtree(test_swc_tree.get_node_list())

    t_matches = {}
    if diadem.g_find_proper_root:
        diadem.adjust_root(
            swc_gold_tree=gold_swc_tree,
            swc_test_tree=test_swc_tree,
            test_kdtree=test_kdtree,
            test_pos_node_dict=test_pos_node_dict,
            t_matches=t_matches,
        )
    if diadem.g_align_tree_by_root:
        test_swc_tree.align_roots(gold_tree=gold_swc_tree, t_matches=t_matches)

    # the id of root here is always -1, so we need to figure out first level sons of root
    for sub_gold_root in gold_swc_tree.root().children:
        if sub_gold_root in t_matches.keys():
            sub_test_root = t_matches[sub_gold_root]
            bin_test_root, bin_test_list = bin_utils.convert_to_binarytrees(sub_test_root)
        else:
            bin_test_root, bin_test_list = None, []

        bin_gold_root, bin_gold_list = bin_utils.convert_to_binarytrees(sub_gold_root)

        if diadem.g_remove_spur > 0:
            diadem.g_spur_set = diadem.remove_spurs(bin_gold_root, 1.0)

        diadem.generate_node_weights(bin_gold_root, bin_gold_list, diadem.g_spur_set)
        diadem.score_trees(bin_gold_root, bin_test_root, bin_gold_list, bin_test_list)

        # debug
        if debug:
            for key in diadem.g_matches:
                print ("match1 = {}, match2 = {}".format(key.data.get_id(), diadem.g_matches[key].data.get_id()))
    diadem.color_tree_only()
    if debug:
        for k in diadem.g_weight_dict:
            print ("id = {} wt = {}".format(k.data.get_id(), diadem.g_weight_dict[k]))

    if config["debug"] is True:
        diadem.print_result()

    res = {
        "weight_sum": diadem.g_weight_sum,
        "score_sum": diadem.g_score_sum,
        "diadem_score": diadem.g_final_score
    }
    return res, gold_swc_tree, test_swc_tree
